# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import errno
import json
import logging
import os.path
import sys
import traceback

from sqlalchemy import Column, create_engine, ForeignKey, Integer, String
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import OperationalError, SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

from sopel.tools import Identifier

if sys.version_info.major >= 3:
    unicode = str
    basestring = str


LOGGER = logging.getLogger(__name__)


def _deserialize(value):
    if value is None:
        return None
    # sqlite likes to return ints for strings that look like ints, even though
    # the column type is string. That's how you do dynamic typing wrong.
    value = unicode(value)
    # Just in case someone's mucking with the DB in a way we can't account for,
    # ignore json parsing errors
    try:
        value = json.loads(value)
    except ValueError:
        pass
    return value


BASE = declarative_base()
MYSQL_TABLE_ARGS = {'mysql_engine': 'InnoDB',
                    'mysql_charset': 'utf8mb4',
                    'mysql_collate': 'utf8mb4_unicode_ci'}


class NickIDs(BASE):
    """Nick IDs table SQLAlchemy class."""
    __tablename__ = 'nick_ids'
    nick_id = Column(Integer, primary_key=True)


class Nicknames(BASE):
    """Nicknames table SQLAlchemy class."""
    __tablename__ = 'nicknames'
    __table_args__ = MYSQL_TABLE_ARGS
    nick_id = Column(Integer, ForeignKey('nick_ids.nick_id'), primary_key=True)
    slug = Column(String(255), primary_key=True)
    canonical = Column(String(255))


class NickValues(BASE):
    """Nick values table SQLAlchemy class."""
    __tablename__ = 'nick_values'
    __table_args__ = MYSQL_TABLE_ARGS
    nick_id = Column(Integer, ForeignKey('nick_ids.nick_id'), primary_key=True)
    key = Column(String(255), primary_key=True)
    value = Column(String(255))


class ChannelValues(BASE):
    """Channel values table SQLAlchemy class."""
    __tablename__ = 'channel_values'
    __table_args__ = MYSQL_TABLE_ARGS
    channel = Column(String(255), primary_key=True)
    key = Column(String(255), primary_key=True)
    value = Column(String(255))


class PluginValues(BASE):
    """Plugin values table SQLAlchemy class."""
    __tablename__ = 'plugin_values'
    __table_args__ = MYSQL_TABLE_ARGS
    plugin = Column(String(255), primary_key=True)
    key = Column(String(255), primary_key=True)
    value = Column(String(255))


class SopelDB(object):
    """Database object class.

    :param config: Sopel's configuration settings
    :type config: :class:`sopel.config.Config`

    This defines a simplified interface for basic, common operations on the
    bot's database. Direct access to the database is also available, to serve
    more complex plugins' needs.

    When configured to use SQLite with a relative filename, the file is assumed
    to be in the directory named by the core setting ``homedir``.

    .. versionadded:: 5.0

    .. versionchanged:: 7.0

        Switched from direct SQLite access to :ref:`SQLAlchemy
        <sqlalchemy:overview>`, allowing users more flexibility around what type
        of database they use (especially on high-load Sopel instances, which may
        run up against SQLite's concurrent-access limitations).

    """

    def __init__(self, config):
        # MySQL - mysql://username:password@localhost/db
        # SQLite - sqlite:////home/sopel/.sopel/default.db
        self.type = config.core.db_type

        # Handle SQLite explicitly as a default
        if self.type == 'sqlite':
            path = config.core.db_filename
            if path is None:
                path = os.path.join(config.core.homedir, config.basename + '.db')
            path = os.path.expanduser(path)
            if not os.path.isabs(path):
                path = os.path.normpath(os.path.join(config.core.homedir, path))
            if not os.path.isdir(os.path.dirname(path)):
                raise OSError(
                    errno.ENOENT,
                    'Cannot create database file. '
                    'No such directory: "{}". Check that configuration setting '
                    'core.db_filename is valid'.format(os.path.dirname(path)),
                    path
                )
            self.filename = path
            self.url = 'sqlite:///%s' % path
        # Otherwise, handle all other database engines
        else:
            query = {}
            if self.type == 'mysql':
                drivername = config.core.db_driver or 'mysql'
                query = {'charset': 'utf8mb4'}
            elif self.type == 'postgres':
                drivername = config.core.db_driver or 'postgresql'
            elif self.type == 'oracle':
                drivername = config.core.db_driver or 'oracle'
            elif self.type == 'mssql':
                drivername = config.core.db_driver or 'mssql+pymssql'
            elif self.type == 'firebird':
                drivername = config.core.db_driver or 'firebird+fdb'
            elif self.type == 'sybase':
                drivername = config.core.db_driver or 'sybase+pysybase'
            else:
                raise Exception('Unknown db_type')

            db_user = config.core.db_user
            db_pass = config.core.db_pass
            db_host = config.core.db_host
            db_port = config.core.db_port  # Optional
            db_name = config.core.db_name  # Optional, depending on DB

            # Ensure we have all our variables defined
            if db_user is None or db_pass is None or db_host is None:
                raise Exception('Please make sure the following core '
                                'configuration values are defined: '
                                'db_user, db_pass, db_host')
            self.url = URL(drivername=drivername, username=db_user,
                           password=db_pass, host=db_host, port=db_port,
                           database=db_name, query=query)

        self.engine = create_engine(self.url, pool_recycle=3600)

        # Catch any errors connecting to database
        try:
            self.engine.connect()
        except OperationalError:
            print("OperationalError: Unable to connect to database.")
            raise

        # Create our tables
        BASE.metadata.create_all(self.engine)

        self.ssession = scoped_session(sessionmaker(bind=self.engine))

    def connect(self):
        """Get a direct database connection.

        :return: a proxied DBAPI connection object; see
                 :meth:`sqlalchemy.engine.Engine.raw_connection()`

        .. important::

           The :attr:`~sopel.config.core_section.CoreSection.db_type` in use
           can change how the raw connection object behaves. You probably want
           to use :meth:`session` and the SQLAlchemy ORM in new plugins, and
           officially support only Sopel 7.0+.

           Note that :meth:`session` is not available in Sopel versions prior
           to 7.0. If your plugin needs to be compatible with older Sopel
           releases, your code *should* use SQLAlchemy via :meth:`session` if
           it is available (Sopel 7.0+) and fall back to direct SQLite access
           via :meth:`connect` if it is not (Sopel 6.x).

           We discourage *publishing* plugins that don't work with all
           supported databases, but you're obviously welcome to take shortcuts
           and support only the engine(s) you need in *private* plugins.

        """
        if self.type != 'sqlite':
            # log non-sqlite uses of raw connections for troubleshooting, since
            # unless the developer had a good reason to use this instead of
            # `session()`, it indicates the plugin was written before Sopel 7.0
            # and might not work right when connected to non-sqlite DBs
            LOGGER.info(
                "Raw connection requested when 'db_type' is not 'sqlite':\n"
                "Consider using 'db.session()' to get a SQLAlchemy session "
                "instead here:\n%s",
                traceback.format_list(traceback.extract_stack()[:-1])[-1][:-1])
        return self.engine.raw_connection()

    def session(self):
        """Get a SQLAlchemy Session object.

        :rtype: :class:`sqlalchemy.orm.session.Session`

        .. versionadded:: 7.0

        .. note::

           If your plugin needs to remain compatible with Sopel versions prior
           to 7.0, you can use :meth:`connect` to get a raw connection. See
           its documentation for relevant warnings and compatibility caveats.

        """
        return self.ssession()

    def execute(self, *args, **kwargs):
        """Execute an arbitrary SQL query against the database.

        :return: the query results
        :rtype: :class:`sqlalchemy.engine.ResultProxy`

        The ``ResultProxy`` object returned is a wrapper around a ``Cursor``
        object as specified by PEP 249.
        """
        return self.engine.execute(*args, **kwargs)

    def get_uri(self):
        """Return a direct URL for the database.

        :return: the database connection URI
        :rtype: str

        This can be used to connect from a plugin using another SQLAlchemy
        instance, for example, without sharing the bot's connection.
        """
        return self.url

    # NICK FUNCTIONS

    def get_nick_id(self, nick, create=True):
        """Return the internal identifier for a given nick.

        :param nick: the nickname for which to fetch an ID
        :type nick: :class:`~sopel.tools.Identifier`
        :param bool create: whether to create an ID if one does not exist
        :raise ValueError: if no ID exists for the given ``nick`` and ``create``
                           is set to ``False``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The nick ID is shared across all of a user's aliases, assuming their
        nicks have been grouped together.

        .. seealso::

            Alias/group management functions: :meth:`alias_nick`,
            :meth:`unalias_nick`, :meth:`merge_nick_groups`, and
            :meth:`delete_nick_group`.

        """
        session = self.ssession()
        slug = nick.lower()
        try:
            nickname = session.query(Nicknames) \
                .filter(Nicknames.slug == slug) \
                .one_or_none()

            if nickname is None:
                # see if it needs case-mapping migration
                nickname = session.query(Nicknames) \
                    .filter(Nicknames.slug == Identifier._lower_swapped(nick)) \
                    .one_or_none()
                if nickname is not None:
                    # it does!
                    nickname.slug = slug
                    session.commit()

            if nickname is None:  # "is /* still */ None", if Python had inline comments
                if not create:
                    raise ValueError('No ID exists for the given nick')
                # Generate a new ID
                nick_id = NickIDs()
                session.add(nick_id)
                session.commit()

                # Create a new Nickname
                nickname = Nicknames(nick_id=nick_id.nick_id, slug=slug, canonical=nick)
                session.add(nickname)
                session.commit()
            return nickname.nick_id
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def alias_nick(self, nick, alias):
        """Create an alias for a nick.

        :param str nick: an existing nickname
        :param str alias: an alias by which ``nick`` should also be known
        :raise ValueError: if the ``alias`` already exists
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To merge two *existing* nick groups, use :meth:`merge_nick_groups`.

            To remove an alias created with this function, use
            :meth:`unalias_nick`.

        """
        nick = Identifier(nick)
        alias = Identifier(alias)
        nick_id = self.get_nick_id(nick)
        session = self.ssession()
        try:
            result = session.query(Nicknames) \
                .filter(Nicknames.slug == alias.lower()) \
                .filter(Nicknames.canonical == alias) \
                .one_or_none()
            if result:
                raise ValueError('Alias already exists.')
            nickname = Nicknames(nick_id=nick_id, slug=alias.lower(), canonical=alias)
            session.add(nickname)
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def set_nick_value(self, nick, key, value):
        """Set or update a value in the key-value store for ``nick``.

        :param str nick: the nickname with which to associate the ``value``
        :param str key: the name by which this ``value`` may be accessed later
        :param mixed value: the value to set for this ``key`` under ``nick``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The ``value`` can be any of a range of types; it need not be a string.
        It will be serialized to JSON before being stored and decoded
        transparently upon retrieval.

        .. seealso::

            To retrieve a value set with this method, use
            :meth:`get_nick_value`.

            To delete a value set with this method, use
            :meth:`delete_nick_value`.

        """
        nick = Identifier(nick)
        value = json.dumps(value, ensure_ascii=False)
        nick_id = self.get_nick_id(nick)
        session = self.ssession()
        try:
            result = session.query(NickValues) \
                .filter(NickValues.nick_id == nick_id) \
                .filter(NickValues.key == key) \
                .one_or_none()
            # NickValue exists, update
            if result:
                result.value = value
                session.commit()
            # DNE - Insert
            else:
                new_nickvalue = NickValues(nick_id=nick_id, key=key, value=value)
                session.add(new_nickvalue)
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def delete_nick_value(self, nick, key):
        """Delete a value from the key-value store for ``nick``.

        :param str nick: the nickname whose values to modify
        :param str key: the name of the value to delete
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To set a value in the first place, use :meth:`set_nick_value`.

            To retrieve a value instead of deleting it, use
            :meth:`get_nick_value`.

        """
        nick = Identifier(nick)
        nick_id = self.get_nick_id(nick)
        session = self.ssession()
        try:
            result = session.query(NickValues) \
                .filter(NickValues.nick_id == nick_id) \
                .filter(NickValues.key == key) \
                .one_or_none()
            # NickValue exists, delete
            if result:
                session.delete(result)
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def get_nick_value(self, nick, key, default=None):
        """Get a value from the key-value store for ``nick``.

        :param str nick: the nickname whose values to access
        :param str key: the name by which the desired value was saved
        :param mixed default: value to return if ``key`` does not have a value
                              set (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        .. seealso::

            To set a value for later retrieval with this method, use
            :meth:`set_nick_value`.

            To delete a value instead of retrieving it, use
            :meth:`delete_nick_value`.

        """
        nick = Identifier(nick)
        session = self.ssession()
        try:
            result = session.query(NickValues) \
                .filter(Nicknames.nick_id == NickValues.nick_id) \
                .filter(Nicknames.slug == nick.lower()) \
                .filter(NickValues.key == key) \
                .one_or_none()
            if result is not None:
                result = result.value
            elif default is not None:
                result = default
            return _deserialize(result)
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def unalias_nick(self, alias):
        """Remove an alias.

        :param str alias: an alias with at least one other nick in its group
        :raise ValueError: if there is not at least one other nick in the group
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To delete an entire group, use :meth:`delete_nick_group`.

            To *add* an alias for a nick, use :meth:`alias_nick`.

        """
        alias = Identifier(alias)
        nick_id = self.get_nick_id(alias, False)
        session = self.ssession()
        try:
            count = session.query(Nicknames) \
                .filter(Nicknames.nick_id == nick_id) \
                .count()
            if count <= 1:
                raise ValueError('Given alias is the only entry in its group.')
            session.query(Nicknames).filter(Nicknames.slug == alias.lower()).delete()
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def delete_nick_group(self, nick):
        """Remove a nickname, all of its aliases, and all of its stored values.

        :param str nick: one of the nicknames in the group to be deleted
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. important::

            This is otherwise known as The Nuclear Option. Be *very* sure that
            you want to do this.

        """
        nick = Identifier(nick)
        nick_id = self.get_nick_id(nick, False)
        session = self.ssession()
        try:
            session.query(Nicknames).filter(Nicknames.nick_id == nick_id).delete()
            session.query(NickValues).filter(NickValues.nick_id == nick_id).delete()
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def merge_nick_groups(self, first_nick, second_nick):
        """Merge two nick groups.

        :param str first_nick: one nick in the first group to merge
        :param str second_nick: one nick in the second group to merge
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        Takes two nicks, which may or may not be registered. Unregistered nicks
        will be registered. Keys which are set for only one of the given nicks
        will be preserved. Where both nicks have values for a given key, the
        value set for the ``first_nick`` will be used.

        A nick group can contain one or many nicknames. Groups containing more
        than one nickname can be created with this function, or by using
        :meth:`alias_nick` to add aliases.

        Note that merging of data only applies to the native key-value store.
        Plugins which define their own tables relying on the nick table will
        need to handle their own merging separately.
        """
        first_id = self.get_nick_id(Identifier(first_nick))
        second_id = self.get_nick_id(Identifier(second_nick))
        session = self.ssession()
        try:
            # Get second_id's values
            res = session.query(NickValues).filter(NickValues.nick_id == second_id).all()
            # Update first_id with second_id values if first_id doesn't have that key
            for row in res:
                first_res = session.query(NickValues) \
                    .filter(NickValues.nick_id == first_id) \
                    .filter(NickValues.key == row.key) \
                    .one_or_none()
                if not first_res:
                    self.set_nick_value(first_nick, row.key, _deserialize(row.value))
            session.query(NickValues).filter(NickValues.nick_id == second_id).delete()
            session.query(Nicknames) \
                .filter(Nicknames.nick_id == second_id) \
                .update({'nick_id': first_id})
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    # CHANNEL FUNCTIONS

    def get_channel_slug(self, chan):
        """Return the case-normalized representation of ``channel``.

        :param str channel: the channel name to normalize, with prefix
                            (required)
        :return str: the case-normalized channel name (or "slug"
                     representation)

        This is useful to make sure that a channel name is stored consistently
        in both the bot's own database and third-party plugins'
        databases/files, without regard for variation in case between
        different clients and/or servers on the network.
        """
        chan = Identifier(chan)
        slug = chan.lower()
        session = self.ssession()
        try:
            count = session.query(ChannelValues) \
                .filter(ChannelValues.channel == slug) \
                .count()

            if count == 0:
                # see if it needs case-mapping migration
                old_rows = session.query(ChannelValues) \
                    .filter(ChannelValues.channel == Identifier._lower_swapped(chan))
                old_count = old_rows.count()
                if old_count > 0:
                    # it does!
                    old_rows.update({ChannelValues.channel: slug})
                    session.commit()

            return slug
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def set_channel_value(self, channel, key, value):
        """Set or update a value in the key-value store for ``channel``.

        :param str channel: the channel with which to associate the ``value``
        :param str key: the name by which this ``value`` may be accessed later
        :param mixed value: the value to set for this ``key`` under ``channel``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The ``value`` can be any of a range of types; it need not be a string.
        It will be serialized to JSON before being stored and decoded
        transparently upon retrieval.

        .. seealso::

            To retrieve a value set with this method, use
            :meth:`get_channel_value`.

            To delete a value set with this method, use
            :meth:`delete_channel_value`.

        """
        channel = self.get_channel_slug(channel)
        value = json.dumps(value, ensure_ascii=False)
        session = self.ssession()
        try:
            result = session.query(ChannelValues) \
                .filter(ChannelValues.channel == channel)\
                .filter(ChannelValues.key == key) \
                .one_or_none()
            # ChannelValue exists, update
            if result:
                result.value = value
                session.commit()
            # DNE - Insert
            else:
                new_channelvalue = ChannelValues(channel=channel, key=key, value=value)
                session.add(new_channelvalue)
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def delete_channel_value(self, channel, key):
        """Delete a value from the key-value store for ``channel``.

        :param str channel: the channel whose values to modify
        :param str key: the name of the value to delete
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To set a value in the first place, use :meth:`set_channel_value`.

            To retrieve a value instead of deleting it, use
            :meth:`get_channel_value`.

        """
        channel = self.get_channel_slug(channel)
        session = self.ssession()
        try:
            result = session.query(ChannelValues) \
                .filter(ChannelValues.channel == channel)\
                .filter(ChannelValues.key == key) \
                .one_or_none()
            # ChannelValue exists, delete
            if result:
                session.delete(result)
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def get_channel_value(self, channel, key, default=None):
        """Get a value from the key-value store for ``channel``.

        :param str channel: the channel whose values to access
        :param str key: the name by which the desired value was saved
        :param mixed default: value to return if ``key`` does not have a value
                              set (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        .. seealso::

            To set a value for later retrieval with this method, use
            :meth:`set_channel_value`.

            To delete a value instead of retrieving it, use
            :meth:`delete_channel_value`.

        """
        channel = self.get_channel_slug(channel)
        session = self.ssession()
        try:
            result = session.query(ChannelValues) \
                .filter(ChannelValues.channel == channel)\
                .filter(ChannelValues.key == key) \
                .one_or_none()
            if result is not None:
                result = result.value
            elif default is not None:
                result = default
            return _deserialize(result)
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    # PLUGIN FUNCTIONS

    def set_plugin_value(self, plugin, key, value):
        """Set or update a value in the key-value store for ``plugin``.

        :param str plugin: the plugin name with which to associate the ``value``
        :param str key: the name by which this ``value`` may be accessed later
        :param mixed value: the value to set for this ``key`` under ``plugin``
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        The ``value`` can be any of a range of types; it need not be a string.
        It will be serialized to JSON before being stored and decoded
        transparently upon retrieval.

        .. seealso::

            To retrieve a value set with this method, use
            :meth:`get_plugin_value`.

            To delete a value set with this method, use
            :meth:`delete_plugin_value`.

        """
        plugin = plugin.lower()
        value = json.dumps(value, ensure_ascii=False)
        session = self.ssession()
        try:
            result = session.query(PluginValues) \
                .filter(PluginValues.plugin == plugin)\
                .filter(PluginValues.key == key) \
                .one_or_none()
            # PluginValue exists, update
            if result:
                result.value = value
                session.commit()
            # DNE - Insert
            else:
                new_pluginvalue = PluginValues(plugin=plugin, key=key, value=value)
                session.add(new_pluginvalue)
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def delete_plugin_value(self, plugin, key):
        """Delete a value from the key-value store for ``plugin``.

        :param str plugin: the plugin name whose values to modify
        :param str key: the name of the value to delete
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. seealso::

            To set a value in the first place, use :meth:`set_plugin_value`.

            To retrieve a value instead of deleting it, use
            :meth:`get_plugin_value`.

        """
        plugin = plugin.lower()
        session = self.ssession()
        try:
            result = session.query(PluginValues) \
                .filter(PluginValues.plugin == plugin)\
                .filter(PluginValues.key == key) \
                .one_or_none()
            # PluginValue exists, update
            if result:
                session.delete(result)
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    def get_plugin_value(self, plugin, key, default=None):
        """Get a value from the key-value store for ``plugin``.

        :param str plugin: the plugin name whose values to access
        :param str key: the name by which the desired value was saved
        :param mixed default: value to return if ``key`` does not have a value
                              set (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        .. seealso::

            To set a value for later retrieval with this method, use
            :meth:`set_plugin_value`.

            To delete a value instead of retrieving it, use
            :meth:`delete_plugin_value`.

        """
        plugin = plugin.lower()
        session = self.ssession()
        try:
            result = session.query(PluginValues) \
                .filter(PluginValues.plugin == plugin)\
                .filter(PluginValues.key == key) \
                .one_or_none()
            if result is not None:
                result = result.value
            elif default is not None:
                result = default
            return _deserialize(result)
        except SQLAlchemyError:
            session.rollback()
            raise
        finally:
            self.ssession.remove()

    # NICK AND CHANNEL FUNCTIONS

    def get_nick_or_channel_value(self, name, key, default=None):
        """Get a value from the key-value store for ``name``.

        :param str name: nick or channel whose values to access
        :param str key: the name by which the desired value was saved
        :param mixed default: value to return if ``key`` does not have a value
                              set (optional)
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        .. versionadded:: 7.0

            The ``default`` parameter.

        This is useful for common logic that is shared between both users and
        channels, as it will fetch the appropriate value based on what type of
        ``name`` it is given.

        .. seealso::

            To get a value for a nick specifically, use :meth:`get_nick_value`.

            To get a value for a channel specifically, use
            :meth:`get_channel_value`.

        """
        name = Identifier(name)
        if name.is_nick():
            return self.get_nick_value(name, key, default)
        else:
            return self.get_channel_value(name, key, default)

    def get_preferred_value(self, names, key):
        """Get a value for the first name which has it set.

        :param list names: a list of channel names and/or nicknames
        :param str key: the name by which the desired value was saved
        :return: the value for ``key`` from the first ``name`` which has it set,
                 or ``None`` if none of the ``names`` has it set
        :raise ~sqlalchemy.exc.SQLAlchemyError: if there is a database error

        This is useful for logic that needs to customize its output based on
        settings stored in the database. For example, it can be used to fall
        back from the triggering user's setting to the current channel's setting
        in case the user has not configured their setting.

        .. note::

            This is the only ``get_*_value()`` method that does not support
            passing a ``default``. Try to avoid using it on ``key``\\s which
            might have ``None`` as a valid value, to avoid ambiguous logic.

        """
        for name in names:
            value = self.get_nick_or_channel_value(name, key)
            if value is not None:
                return value



# coding=utf-8
"""
reload.py - Sopel Plugin Reloader Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import subprocess

from sopel import plugin, plugins


LOGGER = logging.getLogger(__name__)
PLUGIN_OUTPUT_PREFIX = '[reload] '


def _load(bot, plugin):
    # handle errors while loading (if any)
    try:
        plugin.load()
        if plugin.has_setup():
            plugin.setup(bot)
        plugin.register(bot)
    except Exception as e:
        LOGGER.exception('Error loading %s: %s', plugin.name, e)
        raise


@plugin.nickname_command("reload")
@plugin.priority("low")
@plugin.thread(False)
@plugin.require_admin
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def f_reload(bot, trigger):
    """Reloads a plugin (for use by admins only)."""
    name = trigger.group(2)

    if not name or name == '*' or name.upper() == 'ALL THE THINGS':
        bot.reload_plugins()
        bot.say('done')
        return

    if not bot.has_plugin(name):
        bot.reply('"%s" not loaded; try the `load` command.' % name)
        return

    bot.reload_plugin(name)
    plugin_meta = bot.get_plugin_meta(name)
    bot.say('done: %s reloaded (%s from %s)' % (
        name,
        plugin_meta['type'],
        plugin_meta['source'],
    ))


@plugin.nickname_command('update')
@plugin.require_admin
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def f_update(bot, trigger):
    """Pulls the latest versions of all plugins from Git (for use by admins only)."""
    proc = subprocess.Popen('/usr/bin/git pull',
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, shell=True)
    bot.reply(proc.communicate()[0])

    f_reload(bot, trigger)


@plugin.nickname_command("load")
@plugin.priority("low")
@plugin.thread(False)
@plugin.require_admin
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def f_load(bot, trigger):
    """Loads a plugin (for use by admins only)."""
    name = trigger.group(2)
    if not name:
        bot.reply('Load what?')
        return

    if bot.has_plugin(name):
        bot.reply('Plugin already loaded; use the `reload` command.')
        return

    usable_plugins = plugins.get_usable_plugins(bot.config)
    if name not in usable_plugins:
        bot.reply('Plugin %s not found' % name)
        return

    plugin, is_enabled = usable_plugins[name]
    if not is_enabled:
        bot.reply('Plugin %s is disabled' % name)
        return

    try:
        _load(bot, plugin)
    except Exception as error:
        bot.reply('Could not load plugin %s: %s' % (name, error))
    else:
        meta = bot.get_plugin_meta(name)
        bot.say('Plugin %s loaded (%s from %s)' % (
            name,
            meta['type'],
            meta['source'],
        ))


# Catch private messages
@plugin.command("reload")
@plugin.priority("low")
@plugin.thread(False)
@plugin.require_privmsg
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def pm_f_reload(bot, trigger):
    """Wrapper for allowing delivery of .reload command via PM"""
    f_reload(bot, trigger)


@plugin.command('update')
@plugin.require_privmsg
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def pm_f_update(bot, trigger):
    """Wrapper for allowing delivery of .update command via PM"""
    f_update(bot, trigger)


@plugin.command("load")
@plugin.priority("low")
@plugin.thread(False)
@plugin.require_privmsg
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def pm_f_load(bot, trigger):
    """Wrapper for allowing delivery of .load command via PM"""
    f_load(bot, trigger)



# coding=utf-8
"""Tests for message formatting"""
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from sopel.formatting import (
    bold,
    color,
    colors,
    CONTROL_NON_PRINTING,
    CONTROL_NORMAL,
    hex_color,
    italic,
    monospace,
    plain,
    reverse,
    strikethrough,
    underline,
)


def test_color():
    text = 'Hello World'
    assert color(text) == text
    assert color(text, colors.PINK) == '\x0313' + text + '\x03'
    assert color(text, colors.PINK, colors.TEAL) == '\x0313,10' + text + '\x03'
    pytest.raises(ValueError, color, text, 100)
    pytest.raises(ValueError, color, text, 'INVALID')


def test_hex_color():
    text = 'Hello World'
    assert hex_color(text) == text
    assert hex_color(text, '369') == '\x04336699' + text + '\x04'
    assert hex_color(text, '246', '987654') == '\x04224466,987654' + text + '\x04'
    pytest.raises(ValueError, hex_color, text, 0x224466)
    pytest.raises(ValueError, hex_color, text, '1234')
    pytest.raises(ValueError, hex_color, text, 'sixchr')


def test_bold():
    text = 'Hello World'
    assert bold(text) == '\x02' + text + '\x02'


def test_italic():
    text = 'Hello World'
    assert italic(text) == '\x1d' + text + '\x1d'


def test_underline():
    text = 'Hello World'
    assert underline(text) == '\x1f' + text + '\x1f'


def test_strikethrough():
    text = 'Hello World'
    assert strikethrough(text) == '\x1e' + text + '\x1e'


def test_monospace():
    text = 'Hello World'
    assert monospace(text) == '\x11' + text + '\x11'


def test_reverse():
    text = 'Hello World'
    assert reverse(text) == '\x16' + text + '\x16'


def test_plain_color():
    text = 'some text'
    assert plain(color(text, colors.PINK)) == text
    assert plain(color(text, colors.PINK, colors.TEAL)) == text

    tpl = 'b %s a'
    expected = tpl % text
    assert plain(tpl % color(text, colors.PINK)) == expected
    assert plain(tpl % color(text, colors.PINK, colors.TEAL)) == expected


def test_plain_hex_color():
    text = 'some text'
    assert plain(hex_color(text, 'ff0098')) == text
    assert plain(hex_color(text, 'ff0098', '00b571')) == text

    tpl = 'b %s a'
    expected = tpl % text
    assert plain(tpl % hex_color(text, 'ff0098')) == expected
    assert plain(tpl % hex_color(text, 'ff0098', '00b571')) == expected


def test_plain_bold():
    text = 'some text'
    assert plain(bold(text)) == text


def test_plain_italic():
    text = 'some text'
    assert plain(italic(text)) == text


def test_plain_underline():
    text = 'some text'
    assert plain(underline(text)) == text


def test_plain_strikethrough():
    text = 'some text'
    assert plain(strikethrough(text)) == text


def test_plain_monospace():
    text = 'some text'
    assert plain(monospace(text)) == text


def test_plain_reverse():
    text = 'some text'
    assert plain(reverse(text)) == text


def test_plain_reset():
    text = 'some%s text' % CONTROL_NORMAL
    assert plain(text) == 'some text'


@pytest.mark.parametrize('code', CONTROL_NON_PRINTING)
def test_plain_non_printing(code):
    text = 'some%s text' % code
    assert plain(text) == 'some text'


def test_plain_unknown():
    text = 'some \x99text'
    assert plain(text) == text, 'An unknown control code must not be stripped'


def test_plain_emoji():
    text = 'some emoji 💪 in here'
    assert plain(text) == text



# coding=utf-8
"""Tests for core ``sopel.bot`` module"""
from __future__ import absolute_import, division, print_function, unicode_literals

import re

import pytest

from sopel import bot, loader, module, plugin, plugins, trigger
from sopel.plugins import rules
from sopel.tests import rawlist
from sopel.tools import Identifier, SopelMemory


TMP_CONFIG = """
[core]
owner = testnick
nick = TestBot
enable = coretasks
"""

MOCK_MODULE_CONTENT = """# coding=utf-8
import sopel.module


@sopel.module.commands("do")
def command_do(bot, trigger):
    pass


@sopel.module.nickname_commands("info")
def nick_command_info(bot, trigger):
    pass


@sopel.module.action_commands("tell")
def action_command_tell(bot, trigger):
    pass


@sopel.module.interval(5)
def interval5s(bot):
    pass


@sopel.module.interval(10)
def interval10s(bot):
    pass


@sopel.module.url(r'(.+\\.)?example\\.com')
def example_url(bot):
    pass


@sopel.module.rule(r'Hello \\w+')
def rule_hello(bot):
    pass


@sopel.module.event('TOPIC')
def rule_on_topic(bot):
    pass


def shutdown():
    pass


def ignored():
    pass

"""


@pytest.fixture
def tmpconfig(configfactory):
    return configfactory('test.cfg', TMP_CONFIG)


@pytest.fixture
def mockbot(tmpconfig, botfactory):
    return botfactory(tmpconfig)


@pytest.fixture
def mockplugin(tmpdir):
    root = tmpdir.mkdir('loader_mods')
    mod_file = root.join('mockplugin.py')
    mod_file.write(MOCK_MODULE_CONTENT)

    return plugins.handlers.PyFilePlugin(mod_file.strpath)


# -----------------------------------------------------------------------------
# sopel.bot.SopelWrapper

def test_wrapper_say(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.say('Hi!')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :Hi!'
    )


def test_wrapper_say_override_destination(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.say('Hi!', destination='#different')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #different :Hi!'
    )


def test_wrapper_notice(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.notice('Hi!')

    assert mockbot.backend.message_sent == rawlist(
        'NOTICE #channel :Hi!'
    )


def test_wrapper_notice_override_destination(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.notice('Hi!', destination='#different')

    assert mockbot.backend.message_sent == rawlist(
        'NOTICE #different :Hi!'
    )


def test_wrapper_action(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.action('Hi!')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :\x01ACTION Hi!\x01'
    )


def test_wrapper_action_override_destination(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.action('Hi!', destination='#different')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #different :\x01ACTION Hi!\x01'
    )


def test_wrapper_reply(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.reply('Hi!')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :Test: Hi!'
    )


def test_wrapper_reply_override_destination(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.reply('Hi!', destination='#another')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #another :Test: Hi!'
    )


def test_wrapper_reply_override_reply_to(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.reply('Hi!', reply_to='Admin')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :Admin: Hi!'
    )


def test_wrapper_reply_override_destination_reply_to(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.reply('Hi!', destination='#another', reply_to='Admin')

    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #another :Admin: Hi!'
    )


def test_wrapper_kick(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.kick('SpamUser')

    assert mockbot.backend.message_sent == rawlist(
        'KICK #channel SpamUser'
    )


def test_wrapper_kick_message(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.kick('SpamUser', message='Test reason')

    assert mockbot.backend.message_sent == rawlist(
        'KICK #channel SpamUser :Test reason'
    )


def test_wrapper_kick_error_nick(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    with pytest.raises(RuntimeError):
        wrapper.kick(None)

    assert mockbot.backend.message_sent == []


def test_wrapper_kick_error_channel(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG OtherUser :test message')
    with pytest.raises(RuntimeError):
        wrapper.kick('SpamUser')

    assert mockbot.backend.message_sent == []


def test_wrapper_kick_override_destination(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.kick('SpamUser', channel='#another')

    assert mockbot.backend.message_sent == rawlist(
        'KICK #another SpamUser'
    )


def test_wrapper_kick_override_destination_message(mockbot, triggerfactory):
    wrapper = triggerfactory.wrapper(
        mockbot, ':Test!test@example.com PRIVMSG #channel :test message')
    wrapper.kick('SpamUser', channel='#another', message='Test reason')

    assert mockbot.backend.message_sent == rawlist(
        'KICK #another SpamUser :Test reason'
    )


# -----------------------------------------------------------------------------
# Register/Unregister plugins

def test_register_plugin(tmpconfig, mockplugin):
    sopel = bot.Sopel(tmpconfig)
    assert not sopel.has_plugin('mockplugin')

    mockplugin.load()
    mockplugin.setup(sopel)
    mockplugin.register(sopel)

    assert sopel.has_plugin('mockplugin')
    assert sopel.rules.has_command('do')
    assert sopel.rules.has_command('do', plugin='mockplugin')
    assert sopel.rules.has_nick_command('info')
    assert sopel.rules.has_nick_command('info', plugin='mockplugin')
    assert sopel.rules.has_action_command('tell')
    assert sopel.rules.has_action_command('tell', plugin='mockplugin')
    assert sopel.rules.has_url_callback('example_url')
    assert sopel.rules.has_url_callback('example_url', plugin='mockplugin')


def test_register_unregister_plugin(tmpconfig, mockplugin):
    sopel = bot.Sopel(tmpconfig, daemon=False)

    # register the plugin
    mockplugin.load()
    mockplugin.register(sopel)
    assert sopel.has_plugin('mockplugin'), 'The mockplugin must be registered!'

    # unregister it
    mockplugin.unregister(sopel)
    assert not sopel.has_plugin('mockplugin')
    assert not sopel.rules.has_command('do')
    assert not sopel.rules.has_nick_command('info')
    assert not sopel.rules.has_action_command('tell')


def test_remove_plugin_unknown_plugin(tmpconfig):
    sopel = bot.Sopel(tmpconfig, daemon=False)

    handler = plugins.handlers.PyModulePlugin('admin', 'sopel.modules')
    with pytest.raises(plugins.exceptions.PluginNotRegistered):
        sopel.remove_plugin(handler, [], [], [], [])


def test_remove_plugin_unregistered_plugin(tmpconfig):
    sopel = bot.Sopel(tmpconfig, daemon=False)

    # register the plugin
    handler = plugins.handlers.PyModulePlugin('coretasks', 'sopel')
    handler.load()
    handler.register(sopel)

    # Unregister the plugin
    handler.unregister(sopel)

    # And now it must raise an exception
    with pytest.raises(plugins.exceptions.PluginNotRegistered):
        sopel.remove_plugin(handler, [], [], [], [])


def test_reload_plugin_unregistered_plugin(tmpconfig):
    sopel = bot.Sopel(tmpconfig, daemon=False)

    # register the plugin
    handler = plugins.handlers.PyModulePlugin('coretasks', 'sopel')
    handler.load()
    handler.register(sopel)

    # Unregister the plugin
    handler.unregister(sopel)

    # And now it must raise an exception
    with pytest.raises(plugins.exceptions.PluginNotRegistered):
        sopel.reload_plugin(handler.name)


# -----------------------------------------------------------------------------
# register callables, jobs, shutdown, urls

def test_register_callables(tmpconfig):
    sopel = bot.Sopel(tmpconfig)

    @module.rule(r'(hi|hello|hey|sup)')
    def rule_hello(bot, trigger):
        pass

    @plugin.rule_lazy(lambda *args: [re.compile(r'say what')])
    def rule_say_what(bot, trigger):
        pass

    @plugin.find(r'(hi|hello|hey|sup)')
    def rule_find_hello(bot, trigger):
        pass

    @plugin.find_lazy(lambda *args: [re.compile(r'what')])
    def rule_find_what(bot, trigger):
        pass

    @plugin.search(r'(hi|hello|hey|sup)')
    def rule_search_hello(bot, trigger):
        pass

    @plugin.search_lazy(lambda *args: [re.compile(r'what')])
    def rule_search_what(bot, trigger):
        pass

    @module.commands('do')
    @module.example('.do nothing')
    def command_do(bot, trigger):
        """The do command does nothing."""
        pass

    @module.commands('main sub')
    @module.example('.main sub')
    def command_main_sub(bot, trigger):
        """A command with subcommand sub."""
        pass

    @module.commands('main other')
    @module.example('.main other')
    def command_main_other(bot, trigger):
        """A command with subcommand other."""
        pass

    @module.nickname_commands('info')
    @module.example('$nickname: info about this')
    def nick_command_info(bot, trigger):
        """Ask Sopel to get some info about nothing."""
        pass

    @module.action_commands('tell')
    def action_command_tell(bot, trigger):
        pass

    @module.commands('mixed')
    @module.rule('mixing')
    def mixed_rule_command(bot, trigger):
        pass

    @module.event('JOIN')
    @plugin.label('handle_join_event')
    def on_join(bot, trigger):
        pass

    # prepare callables to be registered
    callables = [
        rule_hello,
        rule_say_what,
        rule_find_hello,
        rule_find_what,
        rule_search_hello,
        rule_search_what,
        command_do,
        command_main_sub,
        command_main_other,
        nick_command_info,
        action_command_tell,
        mixed_rule_command,
        on_join,
    ]

    # clean callables and set plugin name by hand
    # since the loader and plugin handlers are excluded here
    for handler in callables:
        loader.clean_callable(handler, tmpconfig)
        handler.plugin_name = 'testplugin'

    # register callables
    sopel.register_callables(callables)

    # trigger rule "hello"
    line = ':Foo!foo@example.com PRIVMSG #sopel :hello'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 3
    assert matches[0][0].get_rule_label() == 'rule_hello'
    assert matches[1][0].get_rule_label() == 'rule_find_hello'
    assert matches[2][0].get_rule_label() == 'rule_search_hello'

    # trigger lazy rule "say what"
    line = ':Foo!foo@example.com PRIVMSG #sopel :say what now?'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 3
    assert matches[0][0].get_rule_label() == 'rule_say_what'
    assert matches[1][0].get_rule_label() == 'rule_find_what'
    assert matches[2][0].get_rule_label() == 'rule_search_what'

    # trigger command "do"
    line = ':Foo!foo@example.com PRIVMSG #sopel :.do'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'do'

    # trigger command with subcommand "main-sub"
    line = ':Foo!foo@example.com PRIVMSG #sopel :.main sub'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'main-sub'

    # trigger command with the other subcommand "main-other"
    line = ':Foo!foo@example.com PRIVMSG #sopel :.main other'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'main-other'

    # trigger nick command "info"
    line = ':Foo!foo@example.com PRIVMSG #sopel :TestBot: info'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'info'

    # trigger action command "tell"
    line = ':Foo!foo@example.com PRIVMSG #sopel :\x01ACTION tell\x01'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'tell'

    # trigger rules with event
    line = ':Foo!foo@example.com JOIN #Sopel'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'handle_join_event'

    # trigger command "mixed"
    line = ':Foo!foo@example.com PRIVMSG #sopel :.mixed'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'mixed'

    # trigger rule "mixed_rule_command"
    line = ':Foo!foo@example.com PRIVMSG #sopel :mixing'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'mixed_rule_command'

    # check documentation
    assert sopel.command_groups == {
        'testplugin': ['do', 'info', 'main other', 'main sub', 'mixed'],
    }

    assert sopel.doc == {
        'do': (
            ['The do command does nothing.'],
            ['.do nothing'],
        ),
        'info': (
            ['Ask Sopel to get some info about nothing.'],
            ['TestBot: info about this'],
        ),
        'main sub': (
            ['A command with subcommand sub.'],
            ['.main sub'],
        ),
        'main other': (
            ['A command with subcommand other.'],
            ['.main other'],
        ),
        'mixed': (
            [],
            [],
        )
    }


def test_register_urls(tmpconfig):
    sopel = bot.Sopel(tmpconfig)

    @module.url(r'https://(\S+)/(.+)?')
    @plugin.label('handle_urls_https')
    def url_callback_https(bot, trigger, match):
        pass

    @module.url(r'http://(\S+)/(.+)?')
    @plugin.label('handle_urls_http')
    def url_callback_http(bot, trigger, match):
        pass

    # prepare callables to be registered
    callables = [
        url_callback_https,
        url_callback_http,
    ]

    # clean callables and set plugin name by hand
    # since the loader and plugin handlers are excluded here
    for handler in callables:
        loader.clean_callable(handler, tmpconfig)
        handler.plugin_name = 'testplugin'

    # register callables
    sopel.register_urls(callables)

    # trigger URL callback "handle_urls_https"
    line = ':Foo!foo@example.com PRIVMSG #sopel :https://example.com/test'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'handle_urls_https'

    # trigger URL callback "handle_urls_https"
    line = ':Foo!foo@example.com PRIVMSG #sopel :http://example.com/test'
    pretrigger = trigger.PreTrigger(sopel.nick, line)

    matches = sopel.rules.get_triggered_rules(sopel, pretrigger)
    assert len(matches) == 1
    assert matches[0][0].get_rule_label() == 'handle_urls_http'


# -----------------------------------------------------------------------------
# call_rule

def test_call_rule(mockbot):
    # setup
    items = []

    def testrule(bot, trigger):
        bot.say('hi')
        items.append(1)
        return "Return Value"

    rule_hello = rules.Rule(
        [re.compile(r'(hi|hello|hey|sup)')],
        plugin='testplugin',
        label='testrule',
        handler=testrule)

    # trigger
    line = ':Test!test@example.com PRIVMSG #channel :hello'
    pretrigger = trigger.PreTrigger(mockbot.nick, line)

    # match
    matches = list(rule_hello.match(mockbot, pretrigger))
    assert len(matches) == 1
    match = matches[0]

    # trigger and wrapper
    rule_trigger = trigger.Trigger(
        mockbot.settings, pretrigger, match, account=None)
    wrapper = bot.SopelWrapper(mockbot, rule_trigger)

    # call rule
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert the rule has been executed
    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :hi'
    )
    assert items == [1]

    # assert the rule is not rate limited
    assert not rule_hello.is_rate_limited(Identifier('Test'))
    assert not rule_hello.is_channel_rate_limited('#channel')
    assert not rule_hello.is_global_rate_limited()

    # call rule again
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert the rule has been executed twice now
    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :hi',
        'PRIVMSG #channel :hi',
    )
    assert items == [1, 1]


def test_call_rule_rate_limited_user(mockbot):
    items = []

    # setup
    def testrule(bot, trigger):
        bot.say('hi')
        items.append(1)
        return "Return Value"

    rule_hello = rules.Rule(
        [re.compile(r'(hi|hello|hey|sup)')],
        plugin='testplugin',
        label='testrule',
        handler=testrule,
        rate_limit=100)

    # trigger
    line = ':Test!test@example.com PRIVMSG #channel :hello'
    pretrigger = trigger.PreTrigger(mockbot.nick, line)

    # match
    matches = list(rule_hello.match(mockbot, pretrigger))
    match = matches[0]

    # trigger and wrapper
    rule_trigger = trigger.Trigger(
        mockbot.settings, pretrigger, match, account=None)
    wrapper = bot.SopelWrapper(mockbot, rule_trigger)

    # call rule
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert the rule has been executed
    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :hi'
    )
    assert items == [1]

    # assert the rule is now rate limited
    assert rule_hello.is_rate_limited(Identifier('Test'))
    assert not rule_hello.is_channel_rate_limited('#channel')
    assert not rule_hello.is_global_rate_limited()

    # call rule again
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert no new message
    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :hi'
    ), 'There must not be any new message sent'
    assert items == [1], 'There must not be any new item'


def test_call_rule_rate_limited_channel(mockbot):
    items = []

    # setup
    def testrule(bot, trigger):
        bot.say('hi')
        items.append(1)
        return "Return Value"

    rule_hello = rules.Rule(
        [re.compile(r'(hi|hello|hey|sup)')],
        plugin='testplugin',
        label='testrule',
        handler=testrule,
        channel_rate_limit=100)

    # trigger
    line = ':Test!test@example.com PRIVMSG #channel :hello'
    pretrigger = trigger.PreTrigger(mockbot.nick, line)

    # match
    matches = list(rule_hello.match(mockbot, pretrigger))
    match = matches[0]

    # trigger and wrapper
    rule_trigger = trigger.Trigger(
        mockbot.settings, pretrigger, match, account=None)
    wrapper = bot.SopelWrapper(mockbot, rule_trigger)

    # call rule
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert the rule is now rate limited
    assert not rule_hello.is_rate_limited(Identifier('Test'))
    assert rule_hello.is_channel_rate_limited('#channel')
    assert not rule_hello.is_global_rate_limited()

    # call rule again
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert no new message
    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :hi'
    ), 'There must not be any new message sent'
    assert items == [1], 'There must not be any new item'


def test_call_rule_rate_limited_global(mockbot):
    items = []

    # setup
    def testrule(bot, trigger):
        bot.say('hi')
        items.append(1)
        return "Return Value"

    rule_hello = rules.Rule(
        [re.compile(r'(hi|hello|hey|sup)')],
        plugin='testplugin',
        label='testrule',
        handler=testrule,
        global_rate_limit=100)

    # trigger
    line = ':Test!test@example.com PRIVMSG #channel :hello'
    pretrigger = trigger.PreTrigger(mockbot.nick, line)

    # match
    matches = list(rule_hello.match(mockbot, pretrigger))
    match = matches[0]

    # trigger and wrapper
    rule_trigger = trigger.Trigger(
        mockbot.settings, pretrigger, match, account=None)
    wrapper = bot.SopelWrapper(mockbot, rule_trigger)

    # call rule
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert the rule is now rate limited
    assert not rule_hello.is_rate_limited(Identifier('Test'))
    assert not rule_hello.is_channel_rate_limited('#channel')
    assert rule_hello.is_global_rate_limited()

    # call rule again
    mockbot.call_rule(rule_hello, wrapper, rule_trigger)

    # assert no new message
    assert mockbot.backend.message_sent == rawlist(
        'PRIVMSG #channel :hi'
    ), 'There must not be any new message sent'
    assert items == [1], 'There must not be any new item'


# -----------------------------------------------------------------------------
# Channel privileges


def test_has_channel_privilege_no_privilege(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # unknown channel
    with pytest.raises(ValueError):
        sopel.has_channel_privilege('#adminchannel', plugin.VOICE)

    # join channel
    server.channel_joined('#adminchannel')

    # check privileges
    assert not sopel.has_channel_privilege(name, plugin.VOICE)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert not sopel.has_channel_privilege(name, plugin.HALFOP)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert not sopel.has_channel_privilege(name, plugin.OP)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert not sopel.has_channel_privilege(name, plugin.ADMIN)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert not sopel.has_channel_privilege(name, plugin.OWNER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert not sopel.has_channel_privilege(name, plugin.OPER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OPER)

    # unknown channel
    with pytest.raises(ValueError):
        sopel.has_channel_privilege('#anotherchannel', plugin.VOICE)


def test_has_channel_privilege_voice(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # join channel
    server.channel_joined('#adminchannel')
    server.mode_set('#adminchannel', '+v', [sopel.nick])

    assert sopel.has_channel_privilege(name, plugin.VOICE)
    assert sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert not sopel.has_channel_privilege(name, plugin.HALFOP)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert not sopel.has_channel_privilege(name, plugin.OP)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert not sopel.has_channel_privilege(name, plugin.ADMIN)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert not sopel.has_channel_privilege(name, plugin.OWNER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert not sopel.has_channel_privilege(name, plugin.OPER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OPER)


def test_has_channel_privilege_halfop(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # join channel
    server.channel_joined('#adminchannel')
    server.mode_set('#adminchannel', '+h', [sopel.nick])

    assert sopel.has_channel_privilege(name, plugin.VOICE)
    assert sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert sopel.has_channel_privilege(name, plugin.HALFOP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert not sopel.has_channel_privilege(name, plugin.OP)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert not sopel.has_channel_privilege(name, plugin.ADMIN)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert not sopel.has_channel_privilege(name, plugin.OWNER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert not sopel.has_channel_privilege(name, plugin.OPER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OPER)


def test_has_channel_privilege_op(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # join channel
    server.channel_joined('#adminchannel')
    server.mode_set('#adminchannel', '+o', [sopel.nick])

    assert sopel.has_channel_privilege(name, plugin.VOICE)
    assert sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert sopel.has_channel_privilege(name, plugin.HALFOP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert sopel.has_channel_privilege(name, plugin.OP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert not sopel.has_channel_privilege(name, plugin.ADMIN)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert not sopel.has_channel_privilege(name, plugin.OWNER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert not sopel.has_channel_privilege(name, plugin.OPER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OPER)


def test_has_channel_privilege_admin(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # join channel
    server.channel_joined('#adminchannel')
    server.mode_set('#adminchannel', '+a', [sopel.nick])

    assert sopel.has_channel_privilege(name, plugin.VOICE)
    assert sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert sopel.has_channel_privilege(name, plugin.HALFOP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert sopel.has_channel_privilege(name, plugin.OP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert sopel.has_channel_privilege(name, plugin.ADMIN)
    assert sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert not sopel.has_channel_privilege(name, plugin.OWNER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert not sopel.has_channel_privilege(name, plugin.OPER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OPER)


def test_has_channel_privilege_owner(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # join channel
    server.channel_joined('#adminchannel')
    server.mode_set('#adminchannel', '+q', [sopel.nick])

    assert sopel.has_channel_privilege(name, plugin.VOICE)
    assert sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert sopel.has_channel_privilege(name, plugin.HALFOP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert sopel.has_channel_privilege(name, plugin.OP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert sopel.has_channel_privilege(name, plugin.ADMIN)
    assert sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert sopel.has_channel_privilege(name, plugin.OWNER)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert not sopel.has_channel_privilege(name, plugin.OPER)
    assert not sopel.has_channel_privilege('#adminchannel', plugin.OPER)


def test_has_channel_privilege_operator(ircfactory, botfactory, tmpconfig):
    sopel = botfactory.preloaded(tmpconfig)
    server = ircfactory(sopel)
    name = Identifier('#adminchannel')

    # join channel
    server.channel_joined('#adminchannel')
    server.mode_set('#adminchannel', '+y', [sopel.nick])

    assert sopel.has_channel_privilege(name, plugin.VOICE)
    assert sopel.has_channel_privilege('#adminchannel', plugin.VOICE)
    assert sopel.has_channel_privilege(name, plugin.HALFOP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.HALFOP)
    assert sopel.has_channel_privilege(name, plugin.OP)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OP)
    assert sopel.has_channel_privilege(name, plugin.ADMIN)
    assert sopel.has_channel_privilege('#adminchannel', plugin.ADMIN)
    assert sopel.has_channel_privilege(name, plugin.OWNER)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OWNER)
    assert sopel.has_channel_privilege(name, plugin.OPER)
    assert sopel.has_channel_privilege('#adminchannel', plugin.OPER)


# -----------------------------------------------------------------------------
# URL Callbacks

def test_search_url_callbacks(tmpconfig):
    """Test search_url_callbacks for a registered URL."""
    sopel = bot.Sopel(tmpconfig, daemon=False)

    def url_handler(*args, **kwargs):
        return None

    sopel.register_url_callback(r'https://example\.com', url_handler)
    results = list(sopel.search_url_callbacks('https://example.com'))
    assert len(results) == 1, 'Expected 1 handler; found %d' % len(results)
    assert url_handler in results[0], 'Once registered, handler must be found'


def test_search_url_callbacks_pattern(tmpconfig):
    """Test search_url_callbacks for a registered regex pattern."""
    sopel = bot.Sopel(tmpconfig, daemon=False)

    def url_handler(*args, **kwargs):
        return None

    sopel.register_url_callback(r'https://(www\.)?example\.com', url_handler)
    results = list(sopel.search_url_callbacks('https://example.com'))
    assert len(results) == 1, 'Expected 1 handler; found %d' % len(results)
    assert url_handler in results[0], 'Once registered, handler must be found'

    results = list(sopel.search_url_callbacks('https://www.example.com'))
    assert len(results) == 1, 'Regex pattern must match both URLs'
    assert url_handler in results[0]


def test_search_url_callbacks_compiled_pattern(tmpconfig):
    """Test search_url_callbacks for a registered compiled regex pattern."""
    sopel = bot.Sopel(tmpconfig, daemon=False)
    url_regex = re.compile(r'https://(www\.)?example\.com')

    def url_handler(*args, **kwargs):
        return None

    sopel.register_url_callback(url_regex, url_handler)
    results = list(sopel.search_url_callbacks('https://example.com'))
    assert len(results) == 1, 'Expected 1 handler; found %d' % len(results)
    assert url_handler in results[0], 'Once registered, handler must be found'

    results = list(sopel.search_url_callbacks('https://www.example.com'))
    assert len(results) == 1, 'Regex pattern must match both URLs'
    assert url_handler in results[0]


def test_search_url_callbacks_not_found(tmpconfig):
    """Test search_url_callbacks when pattern does not match."""
    sopel = bot.Sopel(tmpconfig, daemon=False)
    results = sopel.search_url_callbacks('https://example.com')
    assert not list(results), 'No handler registered; must return an empty list'

    def url_handler(*args, **kwargs):
        return None

    sopel.register_url_callback(r'https://(www\.)?example\.com', url_handler)

    results = sopel.search_url_callbacks('https://not-example.com')
    assert not list(results), 'URL must not match any pattern'


def test_register_url_callback_multiple(tmpconfig):
    """Test register_url_callback replace URL callbacks for a pattern."""
    test_pattern = r'https://(www\.)?example\.com'

    def url_handler(*args, **kwargs):
        return None

    def url_handler_replacement(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)
    sopel.register_url_callback(test_pattern, url_handler)

    results = list(sopel.search_url_callbacks('https://www.example.com'))
    assert url_handler in results[0]

    sopel.register_url_callback(test_pattern, url_handler_replacement)

    results = list(sopel.search_url_callbacks('https://www.example.com'))
    assert len(results) == 1, 'There must be one and only one callback'
    assert url_handler_replacement in results[0], (
        'Handler must have been replaced')


def test_unregister_url_callback(tmpconfig):
    """Test unregister_url_callback removes URL callback for a pattern."""
    test_pattern = r'https://(www\.)?example\.com'

    def url_handler(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)

    # now register a pattern, make sure it still work
    sopel.register_url_callback(test_pattern, url_handler)
    assert list(sopel.search_url_callbacks('https://www.example.com'))

    # unregister this pattern
    sopel.unregister_url_callback(test_pattern, url_handler)

    # now it is not possible to find a callback for this pattern
    results = list(sopel.search_url_callbacks('https://www.example.com'))
    assert not results, 'Unregistered URL callback must not work anymore'


def test_unregister_url_callback_no_memory(tmpconfig):
    """Test unregister_url_callback behavior when bot.memory empty"""
    test_pattern = r'https://(www\.)?example\.com'

    def url_handler(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)
    sopel.unregister_url_callback(test_pattern, url_handler)
    # no exception implies success


# Remove once manual callback management is deprecated (8.0)
def test_unregister_url_callback_manual(tmpconfig):
    """Test unregister_url_callback removes a specific callback that was added manually"""
    test_pattern = r'https://(www\.)?example\.com'

    def url_handler(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)
    sopel.memory["url_callbacks"] = SopelMemory()

    # register a callback manually
    sopel.memory["url_callbacks"][re.compile(test_pattern)] = url_handler
    results = list(sopel.search_url_callbacks("https://www.example.com"))
    assert results[0][0] == url_handler, "Callback must be present"

    # unregister it
    sopel.unregister_url_callback(test_pattern, url_handler)

    results = list(sopel.search_url_callbacks("https://www.example.com"))
    assert not results, "Callback should have been removed"


def test_unregister_url_callback_unknown_pattern(tmpconfig):
    """Test unregister_url_callback pass when pattern is unknown."""
    test_pattern = r'https://(www\.)?example\.com'

    def url_handler(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)

    # now register a pattern, make sure it still work
    sopel.register_url_callback(test_pattern, url_handler)
    assert list(sopel.search_url_callbacks('https://www.example.com'))

    # unregister another pattern (that doesn't exist)
    sopel.unregister_url_callback(r'http://localhost', url_handler)

    # the existing pattern still work
    assert list(sopel.search_url_callbacks('https://www.example.com'))


def test_unregister_url_callback_compiled_pattern(tmpconfig):
    """Test unregister_url_callback works with a compiled regex."""
    test_pattern = r'https://(www\.)?example\.com'
    url_regex = re.compile(test_pattern)

    def url_handler(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)

    # now register a pattern, make sure it still work
    sopel.register_url_callback(test_pattern, url_handler)
    assert list(sopel.search_url_callbacks('https://www.example.com'))

    # unregister using the compiled version
    sopel.unregister_url_callback(url_regex, url_handler)

    assert not list(sopel.search_url_callbacks('https://www.example.com'))


def test_multiple_url_callback(tmpconfig):
    """Test multiple URL callbacks for the same URL."""
    test_pattern_example = r'https://(www\.)?example\.com'
    test_pattern_global = r'https://.*\.com'

    def url_handler(*args, **kwargs):
        return None

    def url_handler_global(*args, **kwargs):
        return None

    sopel = bot.Sopel(tmpconfig, daemon=False)
    sopel.register_url_callback(test_pattern_example, url_handler)
    sopel.register_url_callback(test_pattern_global, url_handler_global)

    results = list(sopel.search_url_callbacks('https://example.com'))
    assert len(results) == 2
    handlers = [result[0] for result in results]

    assert url_handler in handlers
    assert url_handler_global in handlers

    # now unregister one of them: the other must still work
    sopel.unregister_url_callback(test_pattern_example, url_handler)

    results = list(sopel.search_url_callbacks('https://example.com'))
    assert len(results) == 1, 'Exactly one handler must remain'
    assert url_handler_global in results[0], 'Wrong remaining handler'



# coding=utf-8
"""
emoticons.py - Sopel Emoticons Plugin
Copyright 2018, brasstax
Licensed under the Eiffel Forum License 2

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from sopel import plugin


@plugin.command('shrug')
@plugin.action_command('shrugs')
@plugin.example('.shrug', r'¯\_(ツ)_/¯')
def shrug(bot, trigger):
    bot.say('¯\\_(ツ)_/¯')


@plugin.command('happy')
@plugin.example('.happy', 'ᕕ( ᐛ )ᕗ')
def happy(bot, trigger):
    bot.say('ᕕ( ᐛ )ᕗ')


@plugin.command('smirk')
@plugin.action_command('smirks')
@plugin.example('.smirk', '(¬‿¬)')
def smirk(bot, trigger):
    bot.say('(¬‿¬)')


@plugin.command('tableflip', 'tflip')
@plugin.action_command('flips table', 'flips a table', 'flips the table')
@plugin.example('.tableflip', '(╯°□°）╯︵ ┻━┻')
@plugin.example('.tflip', '(╯°□°）╯︵ ┻━┻')
def tableflip(bot, trigger):
    bot.say('(╯°□°）╯︵ ┻━┻')


@plugin.command('unflip')
@plugin.action_command('unflips table', 'unflips the table')
@plugin.example('.unflip', '┬┬ ﻿ノ( ゜-゜ノ)')
def unflip(bot, trigger):
    bot.say('┬┬ ﻿ノ( ゜-゜ノ)')


@plugin.command('lenny')
@plugin.example('.lenny', '( ͡° ͜ʖ ͡°)')
def lenny(bot, trigger):
    bot.say('( ͡° ͜ʖ ͡°)')


@plugin.command('rage', 'anger')
@plugin.example('.rage', 'щ(ಠ益ಠщ)')
@plugin.example('.anger', 'щ(ಠ益ಠщ)')
def anger(bot, trigger):
    bot.say('щ(ಠ益ಠщ)')


@plugin.command('cry')
@plugin.action_command('cries')
@plugin.example('.cry', '( p′︵‵。)')
def cry(bot, trigger):
    bot.say('( p′︵‵。)')


@plugin.command('love')
@plugin.example('.love', '(●♡∀♡)')
def love(bot, trigger):
    bot.say('(●♡∀♡)')


@plugin.command('success', 'winner')
@plugin.example('.success', '٩( ᐛ )و')
@plugin.example('.winner', '٩( ᐛ )و')
def success(bot, trigger):
    bot.say('٩( ᐛ )و')


@plugin.command('confused', 'wat')
@plugin.example('.confused', '(●__●)???')
@plugin.example('.wat', '(●__●)???')
def wat(bot, trigger):
    bot.say('(●__●)???')


@plugin.command('crazy')
@plugin.example('.crazy', '⊙_ʘ')
def crazy(bot, trigger):
    bot.say('⊙_ʘ')


@plugin.command('hungry')
@plugin.example('.hungry', 'ლ(´ڡ`ლ)')
def hungry(bot, trigger):
    bot.say('ლ(´ڡ`ლ)')


@plugin.command('surprised')
@plugin.example('.surprised', '(((( ;°Д°))))')
def surprised(bot, trigger):
    bot.say('(((( ;°Д°))))')


@plugin.command('sick')
@plugin.example('.sick', '(-﹏-。)')
def sick(bot, trigger):
    bot.say('(-﹏-。)')


@plugin.command('afraid')
@plugin.example('.afraid', '(　〇□〇）')
def afraid(bot, trigger):
    bot.say('(　〇□〇）')


@plugin.command('worried')
@plugin.example('.worried', '(　ﾟдﾟ)')
def worried(bot, trigger):
    bot.say('(　ﾟдﾟ)')



# coding=utf-8
"""Core Sopel plugin that handles IRC protocol functions.

This plugin allows the bot to run without user-facing functionality:

* it handles client capability negotiation
* it handles client auth (both nick auth and server auth)
* it handles connection registration (RPL_WELCOME, RPL_LUSERCLIENT), dealing
  with error cases such as nick already in use
* it tracks known channels & users (join, quit, nick change and other events)
* it manages blocked (ignored) users

This is written as a plugin to make it easier to extend to support more
responses to standard IRC codes without having to shove them all into the
dispatch function in :class:`sopel.bot.Sopel` and making it easier to maintain.
"""
# Copyright 2008-2011, Sean B. Palmer (inamidst.com) and Michael Yanovich
# (yanovich.net)
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2012-2015, Elsie Powell embolalia.com
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import base64
import collections
import datetime
import functools
import logging
import re
import sys
import time

from sopel import loader, module, plugin
from sopel.config import ConfigurationError
from sopel.irc import isupport
from sopel.irc.utils import CapReq, MyInfo
from sopel.tools import events, Identifier, iteritems, SopelMemory, target, web


if sys.version_info.major >= 3:
    unicode = str

LOGGER = logging.getLogger(__name__)

CORE_QUERYTYPE = '999'
"""WHOX querytype to indicate requests/responses from coretasks.

Other plugins should use a different querytype.
"""

batched_caps = {}


def setup(bot):
    """Set up the coretasks plugin.

    The setup phase is used to activate the throttle feature to prevent a flood
    of JOIN commands when there are too many channels to join.
    """
    bot.memory['join_events_queue'] = collections.deque()

    # Manage JOIN flood protection
    if bot.settings.core.throttle_join:
        wait_interval = max(bot.settings.core.throttle_wait, 1)

        @module.interval(wait_interval)
        @plugin.label('throttle_join')
        def processing_job(bot):
            _join_event_processing(bot)

        loader.clean_callable(processing_job, bot.settings)
        processing_job.plugin_name = 'coretasks'

        bot.register_jobs([processing_job])


def shutdown(bot):
    """Clean up coretasks-related values in the bot's memory."""
    try:
        bot.memory['join_events_queue'].clear()
    except KeyError:
        pass


def _join_event_processing(bot):
    """Process a batch of JOIN event from the ``join_events_queue`` queue.

    Every time this function is executed, it processes at most
    ``throttle_join`` JOIN events. For each JOIN, it sends a WHO request to
    know more about the channel. This will prevent an excess of flood when
    there are too many channels to join at once.
    """
    batch_size = max(bot.settings.core.throttle_join, 1)
    for _ in range(batch_size):
        try:
            channel = bot.memory['join_events_queue'].popleft()
        except IndexError:
            break
        LOGGER.debug("Sending MODE and WHO after channel JOIN: %s", channel)
        bot.write(["MODE", channel])
        _send_who(bot, channel)


def auth_after_register(bot):
    """Do NickServ/AuthServ auth.

    :param bot: a connected Sopel instance
    :type bot: :class:`sopel.bot.Sopel`

    This function can be used, **after** the bot is connected, to handle one of
    these auth methods:

    * ``nickserv``: send a private message to the NickServ service
    * ``authserv``: send an ``AUTHSERV`` command
    * ``Q``: send an ``AUTH`` command
    * ``userserv``: send a private message to the UserServ service

    .. important::

        If ``core.auth_method`` is set, then ``core.nick_auth_method`` will be
        ignored. If none is set, then this function does nothing.

    """
    if bot.config.core.auth_method:
        auth_method = bot.config.core.auth_method
        auth_username = bot.config.core.auth_username
        auth_password = bot.config.core.auth_password
        auth_target = bot.config.core.auth_target
    elif bot.config.core.nick_auth_method:
        auth_method = bot.config.core.nick_auth_method
        auth_username = (bot.config.core.nick_auth_username or
                         bot.config.core.nick)
        auth_password = bot.config.core.nick_auth_password
        auth_target = bot.config.core.nick_auth_target
    else:
        return

    # nickserv-based auth method needs to check for current nick
    if auth_method == 'nickserv':
        if bot.nick != bot.settings.core.nick:
            LOGGER.warning("Sending nickserv GHOST command.")
            bot.say(
                'GHOST %s %s' % (bot.settings.core.nick, auth_password),
                auth_target or 'NickServ')
        else:
            bot.say('IDENTIFY %s' % auth_password, auth_target or 'NickServ')

    # other methods use account instead of nick
    elif auth_method == 'authserv':
        bot.write(('AUTHSERV', 'auth', auth_username, auth_password))
    elif auth_method == 'Q':
        bot.write(('AUTH', auth_username, auth_password))
    elif auth_method == 'userserv':
        bot.say("LOGIN %s %s" % (auth_username, auth_password),
                auth_target or 'UserServ')


def _execute_perform(bot):
    """Execute commands specified to perform on IRC server connect.

    This function executes the list of commands that can be found in the
    ``core.commands_on_connect`` setting. It automatically replaces any
    ``$nickname`` placeholder in the command with the bot's configured nick.
    """
    if not bot.connection_registered:
        # How did you even get this command, bot?
        raise Exception('Bot must be connected to server to perform commands.')

    commands = bot.config.core.commands_on_connect
    count = len(commands)

    if not count:
        LOGGER.info("No custom command to execute.")
        return

    LOGGER.info("Executing %d custom commands.", count)
    for i, command in enumerate(commands, 1):
        command = command.replace('$nickname', bot.config.core.nick)
        LOGGER.debug("Executing custom command [%d/%d]: %s", i, count, command)
        bot.write((command,))


@plugin.event(events.ERR_NICKNAMEINUSE)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def on_nickname_in_use(bot, trigger):
    """Change the bot's nick when the current one is already in use.

    This can be triggered when the bot disconnects then reconnects before the
    server can notice a client timeout. Other reasons include mischief,
    trolling, and obviously, PEBKAC.

    This will change the current nick by adding a trailing ``_``. If the bot
    sees that a user with its configured nick disconnects (see ``QUIT`` event
    handling), the bot will try to regain it.
    """
    LOGGER.error(
        "Nickname already in use! (Nick: %s; Sender: %s; Args: %r)",
        trigger.nick,
        trigger.sender,
        trigger.args,
    )
    bot.change_current_nick(bot.nick + '_')


@module.require_privmsg("This command only works as a private message.")
@module.require_admin("This command requires admin privileges.")
@module.commands('execute')
def execute_perform(bot, trigger):
    """Execute commands specified to perform on IRC server connect.

    This allows a bot owner or admin to force the execution of commands
    that are automatically performed when the bot connects.
    """
    _execute_perform(bot)


@module.event(events.RPL_WELCOME, events.RPL_LUSERCLIENT)
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def startup(bot, trigger):
    """Do tasks related to connecting to the network.

    ``001 RPL_WELCOME`` is from RFC2812 and is the first message that is sent
    after the connection has been registered on the network.

    ``251 RPL_LUSERCLIENT`` is a mandatory message that is sent after the
    client connects to the server in RFC1459. RFC2812 does not require it and
    all networks might not send it. We support both.

    If ``sopel.irc.AbstractBot.connection_registered`` is set, this function
    does nothing and returns immediately. Otherwise, the flag is set and the
    function proceeds normally to:

    1. trigger auth method
    2. set bot's ``MODE`` (from ``core.modes``)
    3. join channels (or queue them to join later)
    4. check for security when the ``account-tag`` capability is enabled
    5. execute custom commands
    """
    if bot.connection_registered:
        return

    # set flag
    bot.connection_registered = True

    # handle auth method
    auth_after_register(bot)

    # set bot's MODE
    modes = bot.config.core.modes
    if modes:
        if not modes.startswith(('+', '-')):
            # Assume "+" by default.
            modes = '+' + modes
        bot.write(('MODE', bot.nick, modes))

    # join channels
    bot.memory['retry_join'] = SopelMemory()

    channels = bot.config.core.channels
    if not channels:
        LOGGER.info("No initial channels to JOIN.")
    elif bot.config.core.throttle_join:
        throttle_rate = int(bot.config.core.throttle_join)
        throttle_wait = max(bot.config.core.throttle_wait, 1)
        channels_joined = 0

        LOGGER.info(
            "Joining %d channels (with JOIN throttle ON); "
            "this may take a moment.",
            len(channels))

        for channel in channels:
            channels_joined += 1
            if not channels_joined % throttle_rate:
                LOGGER.debug(
                    "Waiting %ds before next JOIN batch.",
                    throttle_wait)
                time.sleep(throttle_wait)
            bot.join(channel)
    else:
        LOGGER.info(
            "Joining %d channels (with JOIN throttle OFF); "
            "this may take a moment.",
            len(channels))

        for channel in bot.config.core.channels:
            bot.join(channel)

    # warn for insecure auth method if necessary
    if (not bot.config.core.owner_account and
            'account-tag' in bot.enabled_capabilities and
            '@' not in bot.config.core.owner):
        msg = (
            "This network supports using network services to identify you as "
            "my owner, rather than just matching your nickname. This is much "
            "more secure. If you'd like to do this, make sure you're logged in "
            "and reply with \"{}useserviceauth\""
        ).format(bot.config.core.help_prefix)
        bot.say(msg, bot.config.core.owner)

    # execute custom commands
    _execute_perform(bot)


@module.event(events.RPL_ISUPPORT)
@module.thread(False)
@module.unblockable
@module.rule('are supported by this server')
@plugin.priority('medium')
def handle_isupport(bot, trigger):
    """Handle ``RPL_ISUPPORT`` events."""
    # remember if NAMESX is known to be supported, before parsing RPL_ISUPPORT
    namesx_support = 'NAMESX' in bot.isupport

    # parse ISUPPORT message from server
    parameters = {}
    for arg in trigger.args:
        try:
            key, value = isupport.parse_parameter(arg)
            parameters[key] = value
        except ValueError:
            # ignore malformed parameter: log a warning and continue
            LOGGER.warning("Unable to parse ISUPPORT parameter: %r", arg)

    bot._isupport = bot._isupport.apply(**parameters)

    # was NAMESX support status updated?
    if not namesx_support and 'NAMESX' in bot.isupport:
        # yes it was!
        if 'multi-prefix' not in bot.server_capabilities:
            # and the server doesn't have the multi-prefix capability
            # so we can ask the server to use the NAMESX feature
            bot.write(('PROTOCTL', 'NAMESX'))


@module.event(events.RPL_MYINFO)
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def parse_reply_myinfo(bot, trigger):
    """Handle ``RPL_MYINFO`` events."""
    # keep <client> <servername> <version> only
    # the trailing parameters (mode types) should be read from ISUPPORT
    bot._myinfo = MyInfo(*trigger.args[0:3])

    LOGGER.info(
        "Received RPL_MYINFO from server: %s, %s, %s",
        bot._myinfo.client,
        bot._myinfo.servername,
        bot._myinfo.version,
    )


@module.require_privmsg()
@module.require_owner()
@module.commands('useserviceauth')
def enable_service_auth(bot, trigger):
    """Set owner's account from an authenticated owner.

    This command can be used to automatically configure ``core.owner_account``
    when the owner is known and has a registered account, but the bot doesn't
    have ``core.owner_account`` configured.

    This doesn't work if the ``account-tag`` capability is not available.
    """
    if bot.config.core.owner_account:
        return
    if 'account-tag' not in bot.enabled_capabilities:
        bot.say('This server does not fully support services auth, so this '
                'command is not available.')
        return
    if not trigger.account:
        bot.say('You must be logged in to network services before using this '
                'command.')
        return
    bot.config.core.owner_account = trigger.account
    bot.config.save()
    bot.say('Success! I will now use network services to identify you as my '
            'owner.')
    LOGGER.info(
        "User %s set %s as owner account.",
        trigger.nick,
        trigger.account,
    )


@module.event(events.ERR_NOCHANMODES)
@plugin.priority('medium')
def retry_join(bot, trigger):
    """Give NickServ enough time to identify on a +R channel.

    Give NickServ enough time to identify, and retry rejoining an
    identified-only (+R) channel. Maximum of ten rejoin attempts.
    """
    channel = trigger.args[1]
    if channel in bot.memory['retry_join'].keys():
        bot.memory['retry_join'][channel] += 1
        if bot.memory['retry_join'][channel] > 10:
            LOGGER.warning("Failed to join %s after 10 attempts.", channel)
            return
        LOGGER.info(
            "Rejoining channel %r failed, will retry in 6s.",
            str(channel))
        time.sleep(6)
    else:
        bot.memory['retry_join'][channel] = 0

    attempt = bot.memory['retry_join'][channel] + 1
    LOGGER.info(
        "Trying to rejoin channel %r (attempt %d/10)",
        str(channel), attempt)
    bot.join(channel)


@module.rule('(.*)')
@module.event(events.RPL_NAMREPLY)
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def handle_names(bot, trigger):
    """Handle NAMES responses.

    This function keeps track of users' privileges when Sopel joins channels.
    """
    names = trigger.split()

    # TODO specific to one channel type. See issue 281.
    channels = re.search(r'(#\S*)', trigger.raw)
    if not channels:
        return
    channel = Identifier(channels.group(1))
    if channel not in bot.privileges:
        bot.privileges[channel] = {}
    if channel not in bot.channels:
        bot.channels[channel] = target.Channel(channel)

    # This could probably be made flexible in the future, but I don't think
    # it'd be worth it.
    # If this ever needs to be updated, remember to change the mode handling in
    # the WHO-handler functions below, too.
    mapping = {
        "+": module.VOICE,
        "%": module.HALFOP,
        "@": module.OP,
        "&": module.ADMIN,
        "~": module.OWNER,
        "!": module.OPER,
    }

    for name in names:
        priv = 0
        for prefix, value in iteritems(mapping):
            if prefix in name:
                priv = priv | value
        nick = Identifier(name.lstrip(''.join(mapping.keys())))
        bot.privileges[channel][nick] = priv
        user = bot.users.get(nick)
        if user is None:
            # It's not possible to set the username/hostname from info received
            # in a NAMES reply, unfortunately.
            # Fortunately, the user should already exist in bot.users by the
            # time this code runs, so this is 99.9% ass-covering.
            user = target.User(nick, None, None)
            bot.users[nick] = user
        bot.channels[channel].add_user(user, privs=priv)


@module.rule('(.*)')
@module.event('MODE')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_modes(bot, trigger):
    """Track changes from received MODE commands."""
    _parse_modes(bot, trigger.args)


@module.priority('high')
@module.event(events.RPL_CHANNELMODEIS)
@module.thread(False)
@module.unblockable
def initial_modes(bot, trigger):
    """Populate channel modes from response to MODE request sent after JOIN."""
    _parse_modes(bot, trigger.args[1:], clear=True)


def _parse_modes(bot, args, clear=False):
    """Parse MODE message and apply changes to internal state."""
    channel_name = Identifier(args[0])
    if channel_name.is_nick():
        # We don't do anything with user modes
        LOGGER.debug("Ignoring user modes: %r", args)
        return

    channel = bot.channels[channel_name]

    # Unreal 3 sometimes sends an extraneous trailing space. If we're short an
    # arg, we'll find out later.
    if args[-1] == "":
        args.pop()
    # If any args are still empty, that's something we may not be prepared for,
    # but let's continue anyway hoping they're trailing / not important.
    if len(args) < 2 or not all(args):
        LOGGER.debug(
            "The server sent a possibly malformed MODE message: %r", args)

    modestring = args[1]
    params = args[2:]

    mapping = {
        "v": module.VOICE,
        "h": module.HALFOP,
        "o": module.OP,
        "a": module.ADMIN,
        "q": module.OWNER,
        "y": module.OPER,
        "Y": module.OPER,
    }

    modes = {}
    if not clear:
        # Work on a copy for some thread safety
        modes.update(channel.modes)

    # Process modes
    sign = ""
    param_idx = 0
    chanmodes = bot.isupport.CHANMODES
    for char in modestring:
        # Are we setting or unsetting
        if char in "+-":
            sign = char
            continue

        if char in chanmodes["A"]:
            # Type A (beI, etc) have a nick or address param to add/remove
            if char not in modes:
                modes[char] = set()
            if sign == "+":
                modes[char].add(params[param_idx])
            elif params[param_idx] in modes[char]:
                modes[char].remove(params[param_idx])
            param_idx += 1
        elif char in chanmodes["B"]:
            # Type B (k, etc) always have a param
            if sign == "+":
                modes[char] = params[param_idx]
            elif char in modes:
                modes.pop(char)
            param_idx += 1
        elif char in chanmodes["C"]:
            # Type C (l, etc) have a param only when setting
            if sign == "+":
                modes[char] = params[param_idx]
                param_idx += 1
            elif char in modes:
                modes.pop(char)
        elif char in chanmodes["D"]:
            # Type D (aciLmMnOpqrRst, etc) have no params
            if sign == "+":
                modes[char] = True
            elif char in modes:
                modes.pop(char)
        elif char in mapping and (
            "PREFIX" not in bot.isupport or char in bot.isupport.PREFIX
        ):
            # User privs modes, always have a param
            nick = Identifier(params[param_idx])
            priv = channel.privileges.get(nick, 0)
            # Log a warning if the two privilege-tracking data structures
            # get out of sync. That should never happen.
            # This is a good place to verify that bot.channels is doing
            # what it's supposed to do before ultimately removing the old,
            # deprecated bot.privileges structure completely.
            ppriv = bot.privileges[channel_name].get(nick, 0)
            if priv != ppriv:
                LOGGER.warning(
                    (
                        "Privilege data error! Please share Sopel's "
                        "raw log with the developers, if enabled. "
                        "(Expected %s == %s for %r in %r)"
                    ),
                    priv,
                    ppriv,
                    nick,
                    channel,
                )
            value = mapping.get(char)
            if value is not None:
                if sign == "+":
                    priv = priv | value
                else:
                    priv = priv & ~value
                bot.privileges[channel_name][nick] = priv
                channel.privileges[nick] = priv
            param_idx += 1
        else:
            # Might be in a mode block past A/B/C/D, but we don't speak those.
            # Send a WHO to ensure no user priv modes we're skipping are lost.
            LOGGER.warning(
                "Unknown MODE message, sending WHO. Message was: %r",
                args,
            )
            _send_who(bot, channel_name)
            return

    if param_idx != len(params):
        LOGGER.warning(
            "Too many arguments received for MODE: args=%r chanmodes=%r",
            args,
            chanmodes,
        )

    channel.modes = modes

    LOGGER.info("Updated mode for channel: %s", channel.name)
    LOGGER.debug("Channel %r mode: %r", str(channel.name), channel.modes)


@module.event('NICK')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_nicks(bot, trigger):
    """Track nickname changes and maintain our chanops list accordingly."""
    old = trigger.nick
    new = Identifier(trigger)

    # Give debug mssage, and PM the owner, if the bot's own nick changes.
    if old == bot.nick and new != bot.nick:
        privmsg = (
            "Hi, I'm your bot, %s. Something has made my nick change. This "
            "can cause some problems for me, and make me do weird things. "
            "You'll probably want to restart me, and figure out what made "
            "that happen so you can stop it happening again. (Usually, it "
            "means you tried to give me a nick that's protected by NickServ.)"
        ) % bot.nick
        debug_msg = (
            "Nick changed by server. This can cause unexpected behavior. "
            "Please restart the bot."
        )
        LOGGER.critical(debug_msg)
        bot.say(privmsg, bot.config.core.owner)
        return

    for channel in bot.privileges:
        channel = Identifier(channel)
        if old in bot.privileges[channel]:
            value = bot.privileges[channel].pop(old)
            bot.privileges[channel][new] = value

    for channel in bot.channels.values():
        channel.rename_user(old, new)
    if old in bot.users:
        bot.users[new] = bot.users.pop(old)

    LOGGER.info("User named %r is now known as %r.", old, str(new))


@module.rule('(.*)')
@module.event('PART')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_part(bot, trigger):
    """Track users leaving channels."""
    nick = trigger.nick
    channel = trigger.sender
    _remove_from_channel(bot, nick, channel)
    LOGGER.info("User %r left a channel: %s", str(nick), channel)


@module.event('KICK')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_kick(bot, trigger):
    """Track users kicked from channels."""
    nick = Identifier(trigger.args[1])
    channel = trigger.sender
    _remove_from_channel(bot, nick, channel)
    LOGGER.info(
        "User %r got kicked by %r from a channel: %s",
        str(nick),
        str(trigger.nick),
        channel,
    )


def _remove_from_channel(bot, nick, channel):
    if nick == bot.nick:
        bot.privileges.pop(channel, None)
        bot.channels.pop(channel, None)

        lost_users = []
        for nick_, user in bot.users.items():
            user.channels.pop(channel, None)
            if not user.channels:
                lost_users.append(nick_)
        for nick_ in lost_users:
            bot.users.pop(nick_, None)
    else:
        bot.privileges[channel].pop(nick, None)

        user = bot.users.get(nick)
        if user and channel in user.channels:
            bot.channels[channel].clear_user(nick)
            if not user.channels:
                bot.users.pop(nick, None)


def _send_who(bot, channel):
    if 'WHOX' in bot.isupport:
        # WHOX syntax, see http://faerion.sourceforge.net/doc/irc/whox.var
        # Needed for accounts in WHO replies. The `CORE_QUERYTYPE` parameter
        # for WHO is used to identify the reply from the server and confirm
        # that it has the requested format. WHO replies with different
        # querytypes in the response were initiated elsewhere and will be
        # ignored.
        bot.write(['WHO', channel, 'a%nuachtf,' + CORE_QUERYTYPE])
    else:
        # We might be on an old network, but we still care about keeping our
        # user list updated
        bot.write(['WHO', channel])
    bot.channels[Identifier(channel)].last_who = datetime.datetime.utcnow()


@module.interval(30)
def _periodic_send_who(bot):
    """Periodically send a WHO request to keep user information up-to-date."""
    if 'away-notify' in bot.enabled_capabilities:
        # WHO not needed to update 'away' status
        return

    # Loops through the channels to find the one that has the longest time since the last WHO
    # request, and issues a WHO request only if the last request for the channel was more than
    # 120 seconds ago.
    who_trigger_time = datetime.datetime.utcnow() - datetime.timedelta(seconds=120)
    selected_channel = None
    for channel_name, channel in bot.channels.items():
        if channel.last_who is None:
            # WHO was never sent yet to this channel: stop here
            selected_channel = channel_name
            break
        if channel.last_who < who_trigger_time:
            # this channel's last who request is the most outdated one at the moment
            selected_channel = channel_name
            who_trigger_time = channel.last_who

    if selected_channel is not None:
        # selected_channel's last who is either none or the oldest valid
        LOGGER.debug("Sending WHO for channel: %s", selected_channel)
        _send_who(bot, selected_channel)


@module.event('JOIN')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_join(bot, trigger):
    """Track users joining channels.

    When a user joins a channel, the bot will send (or queue) a ``WHO`` command
    to know more about said user (privileges, modes, etc.).
    """
    channel = trigger.sender

    # is it a new channel?
    if channel not in bot.channels:
        bot.privileges[channel] = {}
        bot.channels[channel] = target.Channel(channel)

    # did *we* just join?
    if trigger.nick == bot.nick:
        LOGGER.info("Channel joined: %s", channel)
        if bot.settings.core.throttle_join:
            LOGGER.debug("JOIN event added to queue for channel: %s", channel)
            bot.memory['join_events_queue'].append(channel)
        else:
            LOGGER.debug("Send MODE and direct WHO for channel: %s", channel)
            bot.write(["MODE", channel])
            _send_who(bot, channel)
    else:
        LOGGER.info(
            "Channel %r joined by user: %s",
            str(channel), trigger.nick)

    # set initial values
    bot.privileges[channel][trigger.nick] = 0

    user = bot.users.get(trigger.nick)
    if user is None:
        user = target.User(trigger.nick, trigger.user, trigger.host)
        bot.users[trigger.nick] = user
    bot.channels[channel].add_user(user)

    if len(trigger.args) > 1 and trigger.args[1] != '*' and (
            'account-notify' in bot.enabled_capabilities and
            'extended-join' in bot.enabled_capabilities):
        user.account = trigger.args[1]


@module.event('QUIT')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_quit(bot, trigger):
    """Track when users quit channels."""
    for chanprivs in bot.privileges.values():
        chanprivs.pop(trigger.nick, None)
    for channel in bot.channels.values():
        channel.clear_user(trigger.nick)
    bot.users.pop(trigger.nick, None)

    LOGGER.info("User quit: %s", trigger.nick)

    if trigger.nick == bot.settings.core.nick and trigger.nick != bot.nick:
        # old nick is now available, let's change nick again
        bot.change_current_nick(bot.settings.core.nick)
        auth_after_register(bot)


@module.event('CAP')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def receive_cap_list(bot, trigger):
    """Handle client capability negotiation."""
    cap = trigger.strip('-=~')
    # Server is listing capabilities
    if trigger.args[1] == 'LS':
        receive_cap_ls_reply(bot, trigger)
    # Server denied CAP REQ
    elif trigger.args[1] == 'NAK':
        entry = bot._cap_reqs.get(cap, None)
        # If it was requested with bot.cap_req
        if entry:
            for req in entry:
                # And that request was mandatory/prohibit, and a callback was
                # provided
                if req.prefix and req.failure:
                    # Call it.
                    req.failure(bot, req.prefix + cap)
    # Server is removing a capability
    elif trigger.args[1] == 'DEL':
        entry = bot._cap_reqs.get(cap, None)
        # If it was requested with bot.cap_req
        if entry:
            for req in entry:
                # And that request wasn't prohibit, and a callback was
                # provided
                if req.prefix != '-' and req.failure:
                    # Call it.
                    req.failure(bot, req.prefix + cap)
    # Server is adding new capability
    elif trigger.args[1] == 'NEW':
        entry = bot._cap_reqs.get(cap, None)
        # If it was requested with bot.cap_req
        if entry:
            for req in entry:
                # And that request wasn't prohibit
                if req.prefix != '-':
                    # Request it
                    bot.write(('CAP', 'REQ', req.prefix + cap))
    # Server is acknowledging a capability
    elif trigger.args[1] == 'ACK':
        caps = trigger.args[2].split()
        for cap in caps:
            cap.strip('-~= ')
            bot.enabled_capabilities.add(cap)
            entry = bot._cap_reqs.get(cap, [])
            for req in entry:
                if req.success:
                    req.success(bot, req.prefix + trigger)
            if cap == 'sasl':  # TODO why is this not done with bot.cap_req?
                try:
                    receive_cap_ack_sasl(bot)
                except ConfigurationError as error:
                    LOGGER.error(str(error))
                    bot.quit('Wrong SASL configuration.')


def receive_cap_ls_reply(bot, trigger):
    if bot.server_capabilities:
        # We've already seen the results, so someone sent CAP LS from a plugin.
        # We're too late to do SASL, and we don't want to send CAP END before
        # the plugin has done what it needs to, so just return
        return

    for cap in trigger.split():
        c = cap.split('=')
        if len(c) == 2:
            batched_caps[c[0]] = c[1]
        else:
            batched_caps[c[0]] = None

    # Not the last in a multi-line reply. First two args are * and LS.
    if trigger.args[2] == '*':
        return

    LOGGER.info(
        "Client capability negotiation list: %s",
        ', '.join(batched_caps.keys()),
    )
    bot.server_capabilities = batched_caps

    # If some other plugin requests it, we don't need to add another request.
    # If some other plugin prohibits it, we shouldn't request it.
    core_caps = [
        'echo-message',
        'multi-prefix',
        'away-notify',
        'cap-notify',
        'server-time',
    ]
    for cap in core_caps:
        if cap not in bot._cap_reqs:
            bot._cap_reqs[cap] = [CapReq('', 'coretasks')]

    def acct_warn(bot, cap):
        LOGGER.info("Server does not support %s, or it conflicts with a custom "
                    "plugin. User account validation unavailable or limited.",
                    cap[1:])
        if bot.config.core.owner_account or bot.config.core.admin_accounts:
            LOGGER.warning(
                "Owner or admin accounts are configured, but %s is not "
                "supported by the server. This may cause unexpected behavior.",
                cap[1:])
    auth_caps = ['account-notify', 'extended-join', 'account-tag']
    for cap in auth_caps:
        if cap not in bot._cap_reqs:
            bot._cap_reqs[cap] = [CapReq('', 'coretasks', acct_warn)]

    for cap, reqs in iteritems(bot._cap_reqs):
        # At this point, we know mandatory and prohibited don't co-exist, but
        # we need to call back for optionals if they're also prohibited
        prefix = ''
        for entry in reqs:
            if prefix == '-' and entry.prefix != '-':
                entry.failure(bot, entry.prefix + cap)
                continue
            if entry.prefix:
                prefix = entry.prefix

        # It's not required, or it's supported, so we can request it
        if prefix != '=' or cap in bot.server_capabilities:
            # REQs fail as a whole, so we send them one capability at a time
            bot.write(('CAP', 'REQ', entry.prefix + cap))
        # If it's required but not in server caps, we need to call all the
        # callbacks
        else:
            for entry in reqs:
                if entry.failure and entry.prefix == '=':
                    entry.failure(bot, entry.prefix + cap)

    # If we want to do SASL, we have to wait before we can send CAP END. So if
    # we are, wait on 903 (SASL successful) to send it.
    if bot.config.core.auth_method == 'sasl' or bot.config.core.server_auth_method == 'sasl':
        bot.write(('CAP', 'REQ', 'sasl'))
    else:
        bot.write(('CAP', 'END'))
        LOGGER.info("End of client capability negotiation requests.")


def receive_cap_ack_sasl(bot):
    # Presumably we're only here if we said we actually *want* sasl, but still
    # check anyway in case the server glitched.
    password, mech = _get_sasl_pass_and_mech(bot)
    if not password:
        return

    mech = mech or 'PLAIN'
    available_mechs = bot.server_capabilities.get('sasl', '')
    available_mechs = available_mechs.split(',') if available_mechs else []

    if available_mechs and mech not in available_mechs:
        """
        Raise an error if configured to use an unsupported SASL mechanism,
        but only if the server actually advertised supported mechanisms,
        i.e. this network supports SASL 3.2

        SASL 3.1 failure is handled (when possible) by the sasl_mechs() function

        See https://github.com/sopel-irc/sopel/issues/1780 for background
        """
        raise ConfigurationError(
            "SASL mechanism '{}' is not advertised by this server.".format(mech))

    bot.write(('AUTHENTICATE', mech))


def send_authenticate(bot, token):
    """Send ``AUTHENTICATE`` command to server with the given ``token``.

    :param bot: instance of IRC bot that must authenticate
    :param str token: authentication token

    In case the ``token`` is more than 400 bytes, we need to split it and send
    as many ``AUTHENTICATE`` commands as needed. If the last chunk is 400 bytes
    long, we must also send a last empty command (`AUTHENTICATE +` is for empty
    line), so the server knows we are done with ``AUTHENTICATE``.

    .. seealso::

        https://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command

    """
    # payload is a base64 encoded token
    payload = base64.b64encode(token.encode('utf-8'))

    # split the payload into chunks of at most 400 bytes
    chunk_size = 400
    for i in range(0, len(payload), chunk_size):
        offset = i + chunk_size
        chunk = payload[i:offset]
        bot.write(('AUTHENTICATE', chunk))

    # send empty (+) AUTHENTICATE when payload's length is a multiple of 400
    if len(payload) % chunk_size == 0:
        bot.write(('AUTHENTICATE', '+'))


@module.event('AUTHENTICATE')
@plugin.thread(False)
@module.unblockable
@plugin.priority('medium')
def auth_proceed(bot, trigger):
    """Handle client-initiated SASL auth.

    If the chosen mechanism is client-first, the server sends an empty
    response (``AUTHENTICATE +``). In that case, Sopel will handle SASL auth
    that uses a token.

    .. important::

        If ``core.auth_method`` is set, then ``core.server_auth_method`` will
        be ignored. If none is set, then this function does nothing.

    """
    if trigger.args[0] != '+':
        # How did we get here? I am not good with computer.
        return
    # Is this right?
    if bot.config.core.auth_method == 'sasl':
        sasl_username = bot.config.core.auth_username
        sasl_password = bot.config.core.auth_password
    elif bot.config.core.server_auth_method == 'sasl':
        sasl_username = bot.config.core.server_auth_username
        sasl_password = bot.config.core.server_auth_password
    else:
        return
    sasl_username = sasl_username or bot.nick
    sasl_token = _make_sasl_plain_token(sasl_username, sasl_password)
    LOGGER.info("Sending SASL Auth token.")
    send_authenticate(bot, sasl_token)


def _make_sasl_plain_token(account, password):
    return '\x00'.join((account, account, password))


@module.event(events.RPL_SASLSUCCESS)
@plugin.thread(False)
@module.unblockable
@plugin.priority('medium')
def sasl_success(bot, trigger):
    """End CAP request on successful SASL auth.

    If SASL is configured, then the bot won't send ``CAP END`` once it gets
    all the capability responses; it will wait for SASL auth result.

    In this case, the SASL auth is a success, so we can close the negotiation.
    """
    LOGGER.info("Successful SASL Auth.")
    bot.write(('CAP', 'END'))
    LOGGER.info("End of client capability negotiation requests.")


@plugin.event(events.ERR_SASLFAIL)
@plugin.event(events.ERR_SASLTOOLONG)
@plugin.event(events.ERR_SASLABORTED)
@plugin.event(events.ERR_NICKLOCKED)
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def sasl_fail(bot, trigger):
    """SASL Auth Failed: log the error and quit."""
    LOGGER.error(
        "SASL Auth Failed; check your configuration: %s",
        str(trigger))
    bot.quit('SASL Auth Failed')


@module.event(events.RPL_SASLMECHS)
@plugin.thread(False)
@module.unblockable
@plugin.priority('low')
def sasl_mechs(bot, trigger):
    # Presumably we're only here if we said we actually *want* sasl, but still
    # check anyway in case the server glitched.
    password, mech = _get_sasl_pass_and_mech(bot)
    if not password:
        return

    supported_mechs = trigger.args[1].split(',')
    if mech not in supported_mechs:
        """
        How we get here:

        1. Sopel connects to a network advertising SASL 3.1
        2. SASL 3.1 doesn't advertise supported mechanisms up front, so Sopel
           blindly goes ahead with whatever SASL config it's set to use
        3. The server doesn't support the mechanism Sopel used, and is a good
           IRC citizen, so it sends this optional numeric, 908 RPL_SASLMECHS

        Note that misconfigured SASL 3.1 will just silently fail when connected
        to an IRC server NOT implementing the optional 908 reply.

        A network with SASL 3.2 should theoretically never get this far because
        Sopel should catch the unadvertised mechanism in receive_cap_ack_sasl().

        See https://github.com/sopel-irc/sopel/issues/1780 for background
        """
        LOGGER.error(
            "Configured SASL mechanism '%s' is not advertised by this server. "
            "Advertised values: %s",
            mech,
            ', '.join(supported_mechs),
        )
        bot.quit('Wrong SASL configuration.')
    else:
        LOGGER.info(
            "Selected SASL mechanism is %s, advertised: %s",
            mech,
            ', '.join(supported_mechs),
        )


def _get_sasl_pass_and_mech(bot):
    password = None
    mech = None
    if bot.config.core.auth_method == 'sasl':
        password = bot.config.core.auth_password
        mech = bot.config.core.auth_target
    elif bot.config.core.server_auth_method == 'sasl':
        password = bot.config.core.server_auth_password
        mech = bot.config.core.server_auth_sasl_mech
    return password, mech


# Live blocklist editing


@module.commands('blocks')
@module.thread(False)
@module.unblockable
@module.priority('low')
@module.require_admin
def blocks(bot, trigger):
    """
    Manage Sopel's blocking features.\
    See [ignore system documentation]({% link _usage/ignoring-people.md %}).
    """
    STRINGS = {
        "success_del": "Successfully deleted block: %s",
        "success_add": "Successfully added block: %s",
        "no_nick": "No matching nick block found for: %s",
        "no_host": "No matching hostmask block found for: %s",
        "invalid": "Invalid format for %s a block. Try: .blocks add (nick|hostmask) sopel",
        "invalid_display": "Invalid input for displaying blocks.",
        "nonelisted": "No %s listed in the blocklist.",
        'huh': "I could not figure out what you wanted to do.",
    }

    masks = set(s for s in bot.config.core.host_blocks if s != '')
    nicks = set(Identifier(nick)
                for nick in bot.config.core.nick_blocks
                if nick != '')
    text = trigger.group().split()

    if len(text) == 3 and text[1] == "list":
        if text[2] == "hostmask":
            if len(masks) > 0:
                blocked = ', '.join(unicode(mask) for mask in masks)
                bot.say("Blocked hostmasks: {}".format(blocked))
            else:
                bot.reply(STRINGS['nonelisted'] % ('hostmasks'))
        elif text[2] == "nick":
            if len(nicks) > 0:
                blocked = ', '.join(unicode(nick) for nick in nicks)
                bot.say("Blocked nicks: {}".format(blocked))
            else:
                bot.reply(STRINGS['nonelisted'] % ('nicks'))
        else:
            bot.reply(STRINGS['invalid_display'])

    elif len(text) == 4 and text[1] == "add":
        if text[2] == "nick":
            nicks.add(text[3])
            bot.config.core.nick_blocks = nicks
            bot.config.save()
        elif text[2] == "hostmask":
            masks.add(text[3].lower())
            bot.config.core.host_blocks = list(masks)
        else:
            bot.reply(STRINGS['invalid'] % ("adding"))
            return

        bot.reply(STRINGS['success_add'] % (text[3]))

    elif len(text) == 4 and text[1] == "del":
        if text[2] == "nick":
            if Identifier(text[3]) not in nicks:
                bot.reply(STRINGS['no_nick'] % (text[3]))
                return
            nicks.remove(Identifier(text[3]))
            bot.config.core.nick_blocks = [unicode(n) for n in nicks]
            bot.config.save()
            bot.reply(STRINGS['success_del'] % (text[3]))
        elif text[2] == "hostmask":
            mask = text[3].lower()
            if mask not in masks:
                bot.reply(STRINGS['no_host'] % (text[3]))
                return
            masks.remove(mask)
            bot.config.core.host_blocks = [unicode(m) for m in masks]
            bot.config.save()
            bot.reply(STRINGS['success_del'] % (text[3]))
        else:
            bot.reply(STRINGS['invalid'] % ("deleting"))
            return
    else:
        bot.reply(STRINGS['huh'])


@module.event('ACCOUNT')
@plugin.thread(False)
@plugin.unblockable
@plugin.priority('medium')
def account_notify(bot, trigger):
    """Track users' accounts."""
    if trigger.nick not in bot.users:
        bot.users[trigger.nick] = target.User(
            trigger.nick, trigger.user, trigger.host)
    account = trigger.args[0]
    if account == '*':
        account = None
    bot.users[trigger.nick].account = account
    LOGGER.info("Update account for nick %r: %s", trigger.nick, account)


@module.event(events.RPL_WHOSPCRPL)
@plugin.thread(False)
@module.unblockable
@plugin.priority('medium')
def recv_whox(bot, trigger):
    """Track ``WHO`` responses when ``WHOX`` is enabled."""
    if len(trigger.args) < 2 or trigger.args[1] != CORE_QUERYTYPE:
        # Ignored, some plugin probably called WHO
        LOGGER.debug("Ignoring WHO reply for channel '%s'; not queried by coretasks", trigger.args[1])
        return
    if len(trigger.args) != 8:
        LOGGER.warning(
            "While populating `bot.accounts` a WHO response was malformed.")
        return
    _, _, channel, user, host, nick, status, account = trigger.args
    away = 'G' in status
    modes = ''.join([c for c in status if c in '~&@%+!'])
    _record_who(bot, channel, user, host, nick, account, away, modes)


def _record_who(bot, channel, user, host, nick, account=None, away=None, modes=None):
    nick = Identifier(nick)
    channel = Identifier(channel)
    if nick not in bot.users:
        usr = target.User(nick, user, host)
        bot.users[nick] = usr
    else:
        usr = bot.users[nick]
        # check for & fill in sparse User added by handle_names()
        if usr.host is None and host:
            usr.host = host
        if usr.user is None and user:
            usr.user = user
    if account == '0':
        usr.account = None
    else:
        usr.account = account
    if away is not None:
        usr.away = away
    priv = 0
    if modes:
        mapping = {
            "+": module.VOICE,
            "%": module.HALFOP,
            "@": module.OP,
            "&": module.ADMIN,
            "~": module.OWNER,
            "!": module.OPER,
        }
        for c in modes:
            priv = priv | mapping[c]
    if channel not in bot.channels:
        bot.channels[channel] = target.Channel(channel)
    bot.channels[channel].add_user(usr, privs=priv)
    if channel not in bot.privileges:
        bot.privileges[channel] = {}
    bot.privileges[channel][nick] = priv


@module.event(events.RPL_WHOREPLY)
@plugin.thread(False)
@module.unblockable
@plugin.priority('medium')
def recv_who(bot, trigger):
    """Track ``WHO`` responses when ``WHOX`` is not enabled."""
    channel, user, host, _, nick, status = trigger.args[1:7]
    away = 'G' in status
    modes = ''.join([c for c in status if c in '~&@%+!'])
    _record_who(bot, channel, user, host, nick, away=away, modes=modes)


@module.event('AWAY')
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_notify(bot, trigger):
    """Track users going away or coming back."""
    if trigger.nick not in bot.users:
        bot.users[trigger.nick] = target.User(
            trigger.nick, trigger.user, trigger.host)
    user = bot.users[trigger.nick]
    user.away = bool(trigger.args)
    state_change = 'went away' if user.away else 'came back'
    LOGGER.info("User %s: %s", state_change, trigger.nick)


@module.event('TOPIC')
@module.event(events.RPL_TOPIC)
@module.thread(False)
@module.unblockable
@plugin.priority('medium')
def track_topic(bot, trigger):
    """Track channels' topics."""
    if trigger.event != 'TOPIC':
        channel = trigger.args[1]
    else:
        channel = trigger.args[0]
    if channel not in bot.channels:
        return
    bot.channels[channel].topic = trigger.args[-1]
    LOGGER.info("Channel's topic updated: %s", channel)


@module.rule(r'(?u).*(.+://\S+).*')
def handle_url_callbacks(bot, trigger):
    """Dispatch callbacks on URLs

    For each URL found in the trigger, trigger the URL callback registered by
    the ``@url`` decorator.
    """
    schemes = bot.config.core.auto_url_schemes
    # find URLs in the trigger
    for url in web.search_urls(trigger, schemes=schemes):
        # find callbacks for said URL
        for function, match in bot.search_url_callbacks(url):
            # trigger callback defined by the `@url` decorator
            if hasattr(function, 'url_regex'):
                # bake the `match` argument in before passing the callback on
                @functools.wraps(function)
                def decorated(bot, trigger):
                    return function(bot, trigger, match=match)

                bot.call(decorated, bot, trigger)



# coding=utf-8
"""Pytest plugin for Sopel.

.. versionadded:: 7.0
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import re
import sys

import pytest

from sopel import bot, loader, plugins, trigger
from .factories import BotFactory, ConfigFactory, IRCFactory, TriggerFactory, UserFactory


TEMPLATE_TEST_CONFIG = """
[core]
nick = {name}
owner = {owner}
admin = {admin}
"""


def get_disable_setup():
    """Generate a pytest fixture to setup the plugin before running its tests.

    When using ``@example`` for a plugin callable with an expected output,
    pytest will be used to run it as a test. In order to work, this fixture
    must be added to the plugin to set up the plugin before running the test.
    """
    @pytest.fixture(autouse=True)
    def disable_setup(request, monkeypatch):
        setup = getattr(request.module, "setup", None)
        isfixture = hasattr(setup, "_pytestfixturefunction")
        if setup is not None and not isfixture and callable(setup):
            monkeypatch.setattr(
                setup,
                "_pytestfixturefunction",
                pytest.fixture(),
                raising=False,
            )
    return disable_setup


def get_example_test(tested_func, msg, results, privmsg, admin,
                     owner, repeat, use_regexp, ignore=[]):
    """Get a function that calls ``tested_func`` with fake wrapper and trigger.

    :param callable tested_func: a Sopel callable that accepts a
                                 :class:`~.bot.SopelWrapper` and a
                                 :class:`~.trigger.Trigger`
    :param str msg: message that is supposed to trigger the command
    :param list results: expected output from the callable
    :param bool privmsg: if ``True``, make the message appear to have arrived
                         in a private message to the bot; otherwise make it
                         appear to have come from a channel
    :param bool admin: make the message appear to have come from an admin
    :param bool owner: make the message appear to have come from an owner
    :param int repeat: how many times to repeat the test; useful for tests that
                       return random stuff
    :param bool use_regexp: pass ``True`` if ``results`` are in regexp format
    :param list ignore: strings to ignore
    :return: a test function for ``tested_func``
    :rtype: :term:`function`
    """
    def test(configfactory, botfactory, ircfactory):
        test_config = TEMPLATE_TEST_CONFIG.format(
            name='NickName',
            admin=admin,
            owner=owner,
        )
        settings = configfactory('default.cfg', test_config)
        url_schemes = settings.core.auto_url_schemes
        mockbot = botfactory(settings)
        server = ircfactory(mockbot)
        server.channel_joined('#Sopel')

        if not hasattr(tested_func, 'commands'):
            raise AssertionError('Function is not a command.')

        loader.clean_callable(tested_func, settings)
        test_rule = plugins.rules.Command.from_callable(settings, tested_func)
        parse_results = list(test_rule.parse(msg))
        assert parse_results, "Example did not match any command."

        match = parse_results[0]
        sender = mockbot.nick if privmsg else "#channel"
        hostmask = "%s!%s@%s" % (mockbot.nick, "UserName", "example.com")

        # TODO enable message tags
        full_message = ':{} PRIVMSG {} :{}'.format(hostmask, sender, msg)
        pretrigger = trigger.PreTrigger(
            mockbot.nick, full_message, url_schemes=url_schemes)
        test_trigger = trigger.Trigger(mockbot.settings, pretrigger, match)
        pattern = re.compile(r'^%s: ' % re.escape(mockbot.nick))

        # setup module
        module = sys.modules[tested_func.__module__]
        if hasattr(module, 'setup'):
            module.setup(mockbot)

        def isnt_ignored(value):
            """Return True if value doesn't match any re in ignore list."""
            return not any(
                re.match(ignored_line, value)
                for ignored_line in ignore)

        expected_output_count = 0
        for _i in range(repeat):
            expected_output_count += len(results)
            wrapper = bot.SopelWrapper(mockbot, test_trigger)
            tested_func(wrapper, test_trigger)

            output_triggers = (
                trigger.PreTrigger(
                    mockbot.nick,
                    message.decode('utf-8'),
                    url_schemes=url_schemes,
                )
                for message in wrapper.backend.message_sent
            )
            output_texts = (
                # subtract "Sopel: " when necessary
                pattern.sub('', output_trigger.args[-1])
                for output_trigger in output_triggers
            )
            outputs = [text for text in output_texts if isnt_ignored(text)]

            # output length
            assert len(outputs) == expected_output_count

            # output content
            for expected, output in zip(results, outputs):
                if use_regexp:
                    message = (
                        "Output does not match the regex:\n"
                        "Pattern: %s\n"
                        "Output: %s"
                    ) % (expected, output)
                    if not re.match(expected, output):
                        raise AssertionError(message)
                else:
                    assert expected == output

    return test


def insert_into_module(func, module_name, base_name, prefix):
    """Add a function into a module.

    This can be used to add a test function, a setup function, or a fixture
    to an existing module to be used with pytest.
    """
    func.__module__ = module_name
    module = sys.modules[module_name]
    # Make sure the func method does not overwrite anything.
    for i in range(1000):
        func.__name__ = str("%s_%s_%s" % (prefix, base_name, i))
        if not hasattr(module, func.__name__):
            break
    else:
        # 1000 variations of this function's name already exist
        raise RuntimeError('Unable to insert function %s into module %s' % (
            func.__name__, func.__module__
        ))
    setattr(module, func.__name__, func)


@pytest.fixture
def botfactory():
    """Fixture to get a Bot factory.

    :return: a factory to create a mocked bot instance
    :rtype: :class:`sopel.tests.factories.BotFactory`

    This is very useful in unit tests::

        def test_bot(configfactory, botfactory):
            settings = configfactory('... skip for clarity ...')
            bot = botfactory(settings) # no plugins loaded
            # ... do something with the bot

        def test_bot_loaded(configfactory, botfactory):
            settings = configfactory('... skip for clarity ...')
            bot = botfactory.preloaded(settings, ['myplugin'])
            # now the bot has `coretasks` and `myplugin` loaded
    """
    return BotFactory()


@pytest.fixture
def configfactory(tmpdir):
    """Fixture to get a config factory.

    :return: a factory to create test settings
    :rtype: :class:`sopel.tests.factories.ConfigFactory`

    The factory will be automatically configured with a ``tmpdir`` object.
    """
    return ConfigFactory(tmpdir)


@pytest.fixture
def triggerfactory():
    """Fixture to get a trigger factory.

    :return: a factory to create triggers
    :rtype: :class:`sopel.tests.factories.TriggerFactory`
    """
    return TriggerFactory()


@pytest.fixture
def ircfactory():
    """Fixture to get an IRC factory.

    :return: a factory to create mock IRC servers
    :rtype: :class:`sopel.tests.factories.IRCFactory`

    For example, a plugin command could be tested with this::

        from sopel.tests import rawlist

        def test_mycommand(configfactory, botfactory, ircfactory, userfactory):
            settings = configfactory('... skip for clarity ...')
            bot = botfactory(settings, ['myplugin'])
            irc = ircfactory(bot)
            user = userfactory('User')

            irc.say(user, '#test', '.mycommand')

            assert bot.backend.message_sent == rawlist(
                'PRIVMSG #test :My plugin replied this.'
            )
    """
    return IRCFactory()


@pytest.fixture
def userfactory():
    """Fixture to get a user factory.

    :return: a factory to create mock users
    :rtype: :class:`sopel.tests.factories.UserFactory`

    ::

        def test_mycommand(userfactory):
            user = userfactory('User')

            assert user.nick == 'User'
            assert user.user == 'user'
            assert user.host == 'example.com'
            assert user.prefix == 'User!user@example.com'
    """
    return UserFactory()
